"""Spatial signature analysis."""

import warnings

import clustergram
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from pylandstats import settings
from pylandstats.multilandscape import MultiLandscape

__all__ = ["SpatialSignatureAnalysis"]


def _compute_metrics_df(
    landscapes,
    class_metrics,
    classes,
    class_metrics_kwargs,
    class_metrics_fillna,
    landscape_metrics,
    landscape_metrics_kwargs,
):
    flat_metrics_dfs = []
    if len(class_metrics) != 0:
        class_metrics_df = landscapes.compute_class_metrics_df(
            metrics=class_metrics,
            classes=classes,
            metrics_kwargs=class_metrics_kwargs,
            fillna=class_metrics_fillna,
        )
        flat_class_metrics_df = class_metrics_df.unstack(level="class_val")
        flat_class_metrics_df.columns = [
            f"{metric}_{class_val}"
            for metric, class_val in flat_class_metrics_df.columns.values
        ]

        flat_metrics_dfs.append(flat_class_metrics_df)
    if len(landscape_metrics) != 0:
        flat_metrics_dfs.append(
            landscapes.compute_landscape_metrics_df(
                metrics=landscape_metrics, metrics_kwargs=landscape_metrics_kwargs
            )
        )
    return pd.concat(flat_metrics_dfs, axis=1)


def _fit_transform(X, transformer, **transformer_kwargs):
    # ACHTUNG: do not modify X in place to avoid side effects
    _X = transformer(**transformer_kwargs).fit_transform(X)
    if isinstance(X, pd.DataFrame):
        _X = pd.DataFrame(_X, index=X.index, columns=X.columns)

    return _X


class SpatialSignatureAnalysis:
    """Pattern-based analysis of landscapes based on spatial signatures."""

    def __init__(
        self,
        landscapes,
        *,
        class_metrics=None,
        landscape_metrics=None,
        classes=None,
        class_metrics_fillna=None,
        class_metrics_kwargs=None,
        landscape_metrics_kwargs=None,
    ):
        """Initialize the spatial signature analysis.

        Parameters
        ----------
        landscapes : pylandstats.MultiLandscape or list-like of pylandstats.Landscape
            A MultiLandscape object or list of the landscapes to be analyzed.
        class_metrics : list-like of str, optional
            A list-like of strings with the names of the metrics that should be
            computed. If `None`, no class-level metric will be computed.
        classes : list-like, optional
            A list-like of ints or strings with the class values that should be
            considered in the context of this analysis case. If `None` and class-level
            metrics are computed, all unique class values will be considered. Ignored if
            no class-level metrics are computed.
        class_metrics_fillna : bool, optional
            Whether `NaN` values representing landscapes with no occurrences of patches
            of the provided class should be replaced by zero when appropriate, e.g.,
            area and edge metrics (no occurrences mean zero area/edge). If the provided
            value is `None` (default), the value will be taken from
            `settings.CLASS_METRICS_DF_FILLNA`.
        class_metrics_kwargs, landscape_metrics_kwargs : dict, optional
            Dictionary mapping the keyword arguments (values) that should be passed to
            each metric method (key) for the class and landscape-level metrics
            respectively. For instance, to exclude the boundary from the computation of
            `total_edge`, metric_kwargs should map the string 'total_edge' (method name)
            to {'count_boundary': False}. If `None`, each metric will be computed
            according to FRAGSTATS defaults.
        """
        # overall idea: we only store the landscapes (and potentially zone_gser) as
        # attributes. Then, any metric can be computed at any point through the
        # `compute_metrics_df` method. This way, the key landscape attributes are cached
        # so that the cost of computing new metrics is dramatically reduced (almost
        # constant/access time in many cases).
        if isinstance(landscapes, MultiLandscape):
            self.landscapes = landscapes
        else:
            self.landscapes = MultiLandscape(
                landscapes, "landscape_id", np.arange(len(landscapes))
            )
        if hasattr(landscapes, "zone_gser"):
            self.zone_gser = landscapes.zone_gser

        if hasattr(landscapes, "dates"):
            self.dates = landscapes.dates

        if class_metrics is None:
            # class_metrics = Landscape.CLASS_METRICS
            class_metrics = []

        if landscape_metrics is None:
            # landscape_metrics = Landscape.LANDSCAPE_METRICS
            landscape_metrics = []

        self.metrics_df = _compute_metrics_df(
            self.landscapes,
            class_metrics,
            classes,
            class_metrics_kwargs,
            class_metrics_fillna,
            landscape_metrics,
            landscape_metrics_kwargs,
        )

    def decompose(
        self,
        *,
        decomposer=None,
        preprocessor=None,
        preprocessor_kwargs=None,
        imputer=None,
        imputer_kwargs=None,
        **decomposer_kwargs,
    ):
        """Factorize the spatial signature matrix into components.

        Parameters
        ----------
        decomposer : class, optional
            A class that implements the decomposition algorithm. It can be any
            scikit-learn like transformer that implements the `fit`, `transform` and
            `fit_transform` methods and with the `components_` and `n_components`
            attributes. If no value is provided, the default value set in
            `settings.DEFAULT_DECOMPOSER` will be taken.
        preprocessor : class, optional
            A class that implements the preprocessing algorithm. It can be any
            scikit-learn like transformer that implements the `fit_transform` method.
            If no value is provided, the default value set in
            `settings.DEFAULT_PREPROCESSOR` will be taken.
        preprocessor_kwargs : dict, optional
            Keyword arguments to be passed to the initializationof `preprocessor`.
        imputer : class, optional
            A class that implements the imputation algorithm. It can be any scikit-learn
            like transformer that implements the `fit_transform` method. If no value is
            provided, no imputation will be performed.
        imputer_kwargs : dict, optional
            Keyword arguments to be passed to the initialization of `imputer`. Ignored
            if `imputer` is `None`.
        **decomposer_kwargs : dict, optional
            Keyword arguments to be passed to the initialization of `decomposer`.

        Returns
        -------
        components_df : pandas.DataFrame
            A DataFrame with the components of the decomposition. Each row corresponds
            to a landscape and each column to a component.
        decomposer_model : object
            The fitted decomposer model.
        """
        # ACHTUNG: using a copy to avoid modifying the original metrics_df
        X = self.metrics_df.copy()

        if preprocessor is None:
            preprocessor = settings.DEFAULT_PREPROCESSOR

        if preprocessor:  # user can provide `preprocessor=False` to skip this step
            if preprocessor_kwargs is None:
                preprocessor_kwargs = {}
            X = _fit_transform(X, preprocessor, **preprocessor_kwargs)

        if imputer is not None:
            if imputer_kwargs is None:
                imputer_kwargs = {}
            X = _fit_transform(X, imputer, **imputer_kwargs)

        if decomposer is None:
            decomposer = settings.DEFAULT_DECOMPOSER
        try:
            # try if the model accepts nan values
            decompose_model = decomposer(**decomposer_kwargs).fit(X)
        except ValueError:
            warnings.warn(
                "The provided spatial signatures contain NaN values which are not "
                "supported by the decomposition model. In order to proceed, the NaN "
                "values will be dropped. However, you may consider either (i) changing "
                "the chosen metrics or (ii) imputing the NaN values by providing the "
                "`imputer` and `imputer_kwargs` arguments.",
                RuntimeWarning,
            )
            X = X.dropna()
            decompose_model = decomposer(**decomposer_kwargs).fit(X)
        # set X to the reduced matrix but as a data frame with the same index as the
        # original metrics' matrix (taking into account the dropped rows if any)
        return pd.DataFrame(
            decompose_model.transform(X), index=X.index
        ), decompose_model

    def get_loading_df(self, decompose_model, *, columns=None, index=None, **df_kwargs):
        """Get components loadings for each metric.

        Parameters
        ----------
        decompose_model : object
            The decomposition model fitted to the spatial signature matrix.
        columns : list-like, optional
            Column names for the components. If no value is provided, an integer range
            from 0 to `n_components_ - 1` will be used.
        index : list-like, optional
            Index names for the metrics. If no value is provided, the column names of
            `metrics_df` will be used.
        **df_kwargs : dict, optional
            Keyword arguments to be passed to the initialization of `pandas.DataFrame`

        Returns
        -------
        loading_df : pandas.DataFrame
            A DataFrame with the loadings of the components. Each row corresponds to a
            metric and each column to a component.
        """
        if df_kwargs is None:
            _df_kwargs = {}
        else:
            _df_kwargs = df_kwargs.copy()
        if columns is None:
            columns = _df_kwargs.pop(
                "columns",
                range(decompose_model.n_components_),
            )
        if index is None:
            index = _df_kwargs.pop("index", self.metrics_df.columns)
        return pd.DataFrame(
            decompose_model.components_.T, columns=columns, index=index, **_df_kwargs
        )

    def get_cgram(
        self,
        *,
        k_range=None,
        decomposer=None,
        decomposer_kwargs=None,
        preprocessor=None,
        preprocessor_kwargs=None,
        imputer=None,
        imputer_kwargs=None,
        **clustergram_kwargs,
    ):
        """
        Get the clustergram of the spatial signature matrix.

        Parameters
        ----------
        k_range : list-like, optional
            A list-like of integer values with the number of clusters to be tested. If
            no value is provided, the default value set in `settings.DEFAULT_K_RANGE`
            will be taken.
        decomposer : class, optional
            A class that implements the decomposition algorithm. It can be any
            scikit-learn like transformer that implements the `fit`, `transform` and
            `fit_transform` methods and with the `components_` and `n_components`
            attributes. If no value is provided, the clustering algorithm will be
            applied to the original spatial signature matrix.
        decomposer_kwargs : dict, optional
            Keyword arguments to be passed to the initialization of `decomposer`.
            Ignored if `decomposer` is `None`.
        preprocessor : class, optional
            A class that implements the preprocessing algorithm. It can be any
            scikit-learn like transformer that implements the `fit_transform` method.
            If no value is provided, the default value set in
            `settings.DEFAULT_PREPROCESSOR` will be taken.
        preprocessor_kwargs : dict, optional
            Keyword arguments to be passed to the initializationof `preprocessor`.
        imputer : class, optional
            A class that implements the imputation algorithm. It can be any scikit-learn
            like transformer that implements the `fit_transform` method. If no value is
            provided, no imputation will be performed.
        imputer_kwargs : dict, optional
            Keyword arguments to be passed to the initialization of `imputer`. Ignored
            if `imputer` is `None`.
        **clustergram_kwargs : dict, optional
            Keyword arguments to be passed to the initialization of `Clustergram`.

        Returns
        -------
        cgram : clustergram.Clustergram
            The fitted clustergram object.
        """
        # TODO
        # if cluster_traj:
        #     self.flat_metrics_df = self.flat_metrics_df.unstack(level="dates")
        #     self.flat_metrics_df.columns = [
        #         f"{metric}_{year}"
        #         for metric, year in self.flat_metrics_df.columns.values
        #     ]
        if decomposer is not None:
            if decomposer_kwargs is None:
                decomposer_kwargs = {}
            X, _ = self.decompose(
                decomposer=decomposer,
                preprocessor=preprocessor,
                preprocessor_kwargs=preprocessor_kwargs,
                imputer=imputer,
                imputer_kwargs=imputer_kwargs,
                **decomposer_kwargs,
            )
        else:
            # if no decomposer is provided, we can still preprocess and impute the data
            X = self.metrics_df.copy()
            if preprocessor is None:
                preprocessor = settings.DEFAULT_PREPROCESSOR
            if preprocessor_kwargs is None:
                preprocessor_kwargs = {}
            X = _fit_transform(X, preprocessor, **preprocessor_kwargs)
            if imputer is not None:
                if imputer_kwargs is None:
                    imputer_kwargs = {}
                X = _fit_transform(X, imputer, **imputer_kwargs)

        if clustergram_kwargs is None:
            clustergram_kwargs = {}
        if k_range is None:
            # TODO: use settings for the defaults
            # no need to pop k_range from the kwargs since it is already a keyword-only
            # argument of this method
            # k_range = _clustergram_kwargs.pop("k_range", settings.DEFAULT_K_RANGE)
            k_range = settings.DEFAULT_K_RANGE
        try:
            cgram = clustergram.Clustergram(k_range=k_range, **clustergram_kwargs).fit(
                X
            )
        except ValueError:
            warnings.warn(
                "The provided spatial signatures contain NaN values which are not "
                "supported by the clustering model. In order to proceed, the NaN values"
                " will be dropped. However, you may consider either (i) changing the "
                "chosen metrics or (ii) imputing the NaN values by providing the "
                "`imputer` and `imputer_kwargs` arguments.",
                RuntimeWarning,
            )
            X = X.dropna()
            cgram = clustergram.Clustergram(k_range=k_range, **clustergram_kwargs).fit(
                X
            )
        # update cgram.data in place so that it is a data frame (instead of a numpy
        # ndarray) with the same indices and columns as `self.metrics_df`. This ensures
        # that we properly match a cluster label to a landscape (crucial if we have
        # dropped the rows of landscapes with NaN values)
        cgram.data = pd.DataFrame(cgram.data, index=X.index, columns=X.columns)
        return cgram

    @staticmethod
    def scatterplot_cluster_metrics(
        cgram,
        n_clusters,
        metric_x,
        metric_y,
        *,
        ax=None,
        palette_name=None,
        center_marker="x",
        center_plot_kwargs=None,
        **scatterplot_kwargs,
    ):
        """Scatterplot the landscape samples colored by their cluster.

        Parameters
        ----------
        cgram : clustergram.Clustergram
            Fitted clustergram object.
        n_clusters : int
            Number of clusters to use.
        metric_x, metric_y : str
            Strings with the names of the metrics to be plotted on the x and y axes
            respectively. Should be a column of `cgram.data`. Note thus that for
            class-level metrics, the passed string should be of the form
            "{metric}_{class_val}".
        ax : matplotlib.axes.Axes, optional
            Axes object to draw the plot onto, otherwise create a new figure.
        palette_name : str, optional
            Name of palette or None to return current palette.
        center_marker : str, optional, default "x"
            Marker to be used for the cluster centers.
        center_plot_kwargs : dict, optional
            Keyword arguments to plot the cluster centers, which will be passed to
            `matplotlib.axes.Axes.plot`.
        scatterplot_kwargs : dict, optional
            Keyword arguments to be passed to `seaborn.scatterplot`.

        Returns
        -------
        ax : matplotlib.axes.Axes
            Returns the `Axes` object with the plot drawn onto it.

        """
        if scatterplot_kwargs is None:
            _scatterplot_kwargs = {}
        else:
            _scatterplot_kwargs = scatterplot_kwargs.copy()
        if center_plot_kwargs is None:
            _center_plot_kwargs = {}
        else:
            _center_plot_kwargs = center_plot_kwargs.copy()
            center_marker = _center_plot_kwargs.pop("marker", center_marker)
        _center_plot_kwargs["marker"] = center_marker
        if ax is None:
            ax = _scatterplot_kwargs.pop("ax", None)
        if ax is None:
            _, ax = plt.subplots()

        cluster_centers = cgram.cluster_centers[n_clusters][
            :,
            [cgram.data.columns.get_loc(metric) for metric in [metric_x, metric_y]],
        ]
        n_clusters = cluster_centers.shape[0]
        palette = _scatterplot_kwargs.pop(
            "palette", sns.color_palette(palette_name, n_colors=n_clusters)
        )
        sns.scatterplot(
            x=metric_x,
            y=metric_y,
            hue=pd.Series(cgram.labels_[n_clusters], name="cluster").values,
            data=cgram.data,
            palette=palette,
            ax=ax,
            **_scatterplot_kwargs,
        )

        for color, cluster_center in zip(palette, cluster_centers):
            ax.plot(
                cluster_center[0],
                cluster_center[1],
                color=color,
                **_center_plot_kwargs,
            )

        return ax

    def plot_cluster_landscapes(
        self,
        cgram,
        n_clusters,
        *,
        n_cluster_landscapes=4,
        n_cols=4,
        sample_kwargs=None,
        figure_kwargs=None,
        subfigures_kwargs=None,
        subplots_kwargs=None,
        supylabel_kwargs=None,
        **plot_landscape_kwargs,
    ):
        """Scatterplot the landscape samples colored by their cluster.

        Parameters
        ----------
        cgram : Clustergram
            Fitted clustergram object.
        n_clusters : int
            Number of clusters to use.
        n_cluster_landscapes : int, optional, default 4
            Number of landscapes to plot for each cluster. Providing a value of None
            will plot all landscapes.
        n_cols : int, optional, default 4
            Number of columns for the figure.
        sample_kwargs, figure_kwargs, subfigures_kwargs, subplots_kwargs, \
            supylabel_kwargs, plot_landscape_kwargs : dict, optional
            Keyword arguments to be passed to `pandas.DataFrame.sample`,
            `matplotlib.pyplot.figure`,
            `matplotlib.figure.Figure.subfigures`,
            `matplotlib.figure.SubFigure.subplots`,
            `matplotlib.figure.SubFigure.supylabel` and
            `pylandstats.Landscape.plot_landscape` respectively.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure with its corresponding plots drawn into its axes.
        """
        if figure_kwargs is None:
            _figure_kwargs = {}
        else:
            _figure_kwargs = figure_kwargs.copy()
        if subfigures_kwargs is None:
            subfigures_kwargs = {}
        if subplots_kwargs is None:
            subplots_kwargs = {}
        if supylabel_kwargs is None:
            supylabel_kwargs = {}
        if plot_landscape_kwargs is None:
            _plot_landscape_kwargs = {}
        else:
            _plot_landscape_kwargs = plot_landscape_kwargs.copy()
            _ = _plot_landscape_kwargs.pop("ax", None)

        labels = cgram.labels_[n_clusters]
        metrics_df = self.metrics_df.loc[cgram.data.index]
        if n_cluster_landscapes is None:
            # plot all landscapes
            n_rows_per_cluster = np.ceil(
                np.unique(labels, return_counts=True)[1] / n_cols
            ).astype(int)
        else:
            n_rows_per_cluster = np.full(
                n_clusters, np.ceil(n_cluster_landscapes / n_cols), dtype=int
            )
            if sample_kwargs is None:
                _sample_kwargs = {}
            else:
                _sample_kwargs = sample_kwargs.copy()
                _ = _sample_kwargs.pop("frac", None)
            # use sample to shuffle then groupby and head (because head does not raise
            # an error if the group has less elements than n_cluster_landscapes)
            metrics_df = (
                self.metrics_df.loc[cgram.data.index]
                .sample(frac=1, **_sample_kwargs)
                .groupby(labels)
                .head(n_cluster_landscapes)
            )
            # update labels to match the new metrics_df
            labels = pd.Series(labels, index=cgram.data.index)[metrics_df.index].astype(
                int
            )

        n_rows = n_rows_per_cluster.sum()

        figsize = _figure_kwargs.pop("figsize", None)
        if figsize is None:
            figwidth, figheight = plt.rcParams["figure.figsize"]
            figsize = (figwidth * n_cols, figheight * n_rows)
        fig = plt.figure(figsize=figsize, **_figure_kwargs)
        # gs = fig.add_gridspec(n_rows, n_cols)
        subfigs = fig.subfigures(
            n_clusters,
            1,
            height_ratios=n_rows_per_cluster,
            **subfigures_kwargs,
        )

        # TODO: manage nan metrics better in case or reduction or not
        for subfig, n_cluster_rows, (cluster_label, _cluster_df) in zip(
            subfigs,
            n_rows_per_cluster,
            metrics_df.groupby(labels.values),
        ):
            # n_cluster_rows = np.ceil(len(_cluster_df) / n_cols).astype(int)
            axes = subfig.subplots(n_cluster_rows, n_cols, **subplots_kwargs)
            try:
                axes = axes.flat
            except AttributeError:
                axes = [axes]
            subfig.supylabel(f"Cluster {cluster_label}", **supylabel_kwargs)

            for ls_i, ax in zip(_cluster_df.index, axes):
                self.landscapes.landscape_ser[ls_i].plot_landscape(
                    ax=ax, **plot_landscape_kwargs
                )
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)

            # if the cluster does not fill all the columns of the last row, set the
            # respective axes off
            n_empty_axes = len(axes) - len(_cluster_df) % n_cols
            if n_empty_axes > 0:
                for ax in axes[-n_empty_axes:]:
                    ax.set_axis_off()

        return fig

    def plot_cluster_zones(
        self,
        cgram,
        n_clusters,
        *,
        legend=True,
        categorical=True,
        ax=None,
        subplots_kwargs=None,
        **plot_kwargs,
    ):
        """Plot the landscape zones colored by their cluster.

        Only available if the `SpatialSignatureAnalysis` object has been initialized
        with a zonal analysis class (or subclass). If the `SpatialSignatureAnalysis`
        object has been initialized with a spatio-temporal zonal analysis class (or
        subclass), each date snapshot will be plotted separately.

        Parameters
        ----------
        cgram : Clustergram
            Fitted clustergram object.
        n_clusters : int
            Number of clusters to use.
        ax : matplotlib.axes.Axes, optional
            Axes object to draw the plot onto, otherwise create a new figure. Ignored
            if the `SpatialSignatureAnalysis` object has been initialized with a
            spatio-temporal zonal analysis class (or subclass).
        legend : bool, optional, default True
            Whether to show the legend.
        categorical : bool, optional, default True
            Whether the cluster color map should be categorical.
        subplots_kwargs, plot_kwargs : dict, default None
            Keyword arguments to be passed to `matplotlib.pyplot.subplots` and
            `geopandas.GeoDataFrame.plot` respectively.

        Returns
        -------
        matplotlib.figure.Figure or matplotlib.axes.Axes
            If the `SpatialSignatureAnalysis` object has been initialized with a
            spatio-temporal zonal analysis class (or subclass), returns the figure with
            its corresponding plots drawn into its axes. Otherwise, returns the axes
            object with the plot drawn onto it.
        """
        if not hasattr(self, "zone_gser"):
            warnings.warn(
                "The `plot_cluster_zones` method can only be called when"
                " `SpatialSignatureAnalysis` has been initialized with a zonal analysis"
                " class (or sub-class) as the `landscapes` argument. Returning None",
                RuntimeWarning,
            )
            return None

        def _plot(cluster_labels, cgram_data, ax):
            return gpd.GeoDataFrame(
                {"cluster": cluster_labels},
                geometry=cgram_data.index.map(self.zone_gser).values,
            ).plot(
                "cluster", legend=legend, categorical=categorical, ax=ax, **plot_kwargs
            )

        if ax is None or hasattr(self, "dates"):
            if subplots_kwargs is None:
                _subplots_kwargs = {}
            else:
                _subplots_kwargs = subplots_kwargs.copy()

            if hasattr(self, "dates"):
                figsize = _subplots_kwargs.pop(
                    "figsize",
                    (
                        plt.rcParams["figure.figsize"][0] * len(self.dates),
                        plt.rcParams["figure.figsize"][1],
                    ),
                )
                fig, axes = plt.subplots(
                    1, len(self.dates), figsize=figsize, **_subplots_kwargs
                )
                for date, ax in zip(self.dates, axes):
                    date_query = cgram.data.index.get_level_values("date") == date
                    _plot(
                        cgram.labels_[n_clusters][date_query],
                        cgram.data[date_query].droplevel("date"),
                        ax,
                    )
                    ax.set_title(date)
                return fig
            elif ax is None:
                figsize = _subplots_kwargs.pop(
                    "figsize", plt.rcParams["figure.figsize"]
                )
                _, ax = plt.subplots(figsize=figsize, **_subplots_kwargs)

        return _plot(cgram.labels_[n_clusters], cgram.data, ax)



"""Zonal analysis."""

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio as rio
from rasterio import features, mask
from shapely import geometry
from shapely.geometry import base as geometry_base

try:
    from fiona import errors as fiona_errors
except ImportError:
    fiona_errors = None
try:
    from pyogrio import errors as pyogrio_errors
except ImportError:
    pyogrio_errors = None


from pylandstats import multilandscape
from pylandstats.landscape import Landscape

__all__ = ["ZonalAnalysis", "BufferAnalysis", "ZonalGridAnalysis"]

ZONES_READ_ERRORS = [AttributeError, TypeError]
if fiona_errors is not None:
    ZONES_READ_ERRORS.append(fiona_errors.DriverError)
if pyogrio_errors is not None:
    ZONES_READ_ERRORS.append(pyogrio_errors.DataSourceError)
ZONES_READ_ERRORS = tuple(ZONES_READ_ERRORS)

_compute_zonal_statistics_gdf_doc = """
Compute the zonal statistics geo-data frame over the landscape raster.

Parameters
----------
metrics : list-like, optional
    A list-like of strings with the names of the metrics that should be computed. If
    `None`, all the implemented metrics at the specified level will be computed.
level : {{'class', 'landscape'}}, optional
    Whether the metrics should be computed at the class or landscape level. If `None`,
    the metrics will be computed (a) at the class level when a non-None `classes` is
    provided, otherwise (b) at the landscape level.
class_val : int, optional
    If provided, the metric will be computed at the level of the corresponding class,
    otherwise it will be computed at the landscape level.
metrics_kwargs : dict, optional
    Dictionary mapping the keyword arguments (values) that should be passed to
    each metric method (key), e.g., to exclude the boundary from the computation
    of `total_edge`, metric_kwargs should map the string 'total_edge' (method name)
    to {{'count_boundary': False}}. If `None`, each metric will be computed
    according to FRAGSTATS defaults.

Returns
-------
zonal_statistics_gdf : geopandas.GeoDataFrame
    Geo-data frame with the computed zonal statistics, with the zones as rows and
    {col_return} as columns.
"""


class ZonalAnalysis(multilandscape.MultiLandscape):
    """Zonal analysis."""

    def __init__(
        self,
        landscape_filepath,
        zones,
        *,
        zone_index=None,
        zone_nodata=0,
        neighborhood_rule=None,
    ):
        """Initialize the zonal analysis.

        Parameters
        ----------
        landscape_filepath : str, file-like object or pathlib.Path object
            A string/file-like object/pathlib.Path object with the landscape data.
        zones : geopandas.GeoSeries, geopandas.GeoDataFrame, list-like, str, \
            file-like object, pathlib.Path object, numpy.ndarray, optional
            This can either be:

            * A geopandas geo-series or geo-data frame.
            * A list-like of shapely geometries, in the CRS of the landscape.
            * A filename or URL, a file-like object opened in binary ('rb') mode, or a
              Path object that will be passed to `geopandas.read_file`.
            * A numpy array of the same shape as the landscape raster image, where each
              zone is labelled by a unique integer value. The values will be used to
              identify the zones (i.e., as index) - unless a different `zone_index` is
              provided.
        zone_index : list-like or str, optional
            Index used to identify zones (i.e., index of the metrics data frames). This
            can either be:

            * A list-like of index labels that will be positionally mapped to each zone.
            * A str with the name of a column only when `zones` is a geo-data frame or a
              geo-data frame file, e.g., a shapefile.
        zone_nodata : numeric, optional, default 0
            Value of the `zones` array that corresponds to no data. Only considered if
            `zones` is a numpy array of integer types.
        neighborhood_rule : {'8', '4'}, optional
            Neighborhood rule to determine patch adjacencies, i.e: '8' (queen's
            case/Moore neighborhood) or '4' (rook's case/Von Neumann neighborhood).
            Ignored if `landscape` is a `Landscape` instance. If no value is provided,
            the default value set in `settings.DEFAULT_NEIGHBORHOOD_RULE` will be taken.
        """
        # the overall approach to process the `zones` argument is: unless `zones` is
        # provided as a geo-series or a list-like of shapely geometries, we first
        # convert it to a geo-data frame, process it (mainly try to get the proper
        # index) and then convert it to a geo-series to store it as instance attribute

        # first, try to read the `zones` argument as a geo-data frame-like file
        try:
            zones = gpd.read_file(zones)
        except ZONES_READ_ERRORS:
            # Depending on the system and installed libraries, geopandas may raise an
            # `AttributeError`, `TypeError` or `fiona.errors.DriverError`. we let this
            # continue and try to read the `zones` argument differently below
            pass

        with rio.open(landscape_filepath) as src:
            if isinstance(zones, np.ndarray):
                # zones is an ndarray labelling each zone by a unique integer value
                # we first instantiate a geo-data frame because we will use the labels
                # as zone ids
                zones = gpd.GeoDataFrame(
                    [
                        (val, geometry.shape(geom))
                        for geom, val in features.shapes(zones, transform=src.transform)
                        if val != zone_nodata
                    ],
                    columns=["zone", "geometry"],
                    crs=src.crs,
                )
                if zone_index is None:
                    # if no zone indexing is provided, we will use the zone labels as
                    # index
                    # zones = zones.set_index("zone")
                    # instead of using `set_index`, we will just set `zone_index` to the
                    # column name "zone", so that the `set_index` is called when
                    # processing the geo-data frame below
                    zone_index = "zone"

            # at this point, unless `zones` was provided as geo series or list-like of
            # shapely geometries, `zones` must be a geo-data frame
            if isinstance(zones, gpd.GeoDataFrame):
                # if there is a non-None `zone_index`, use it
                if zone_index is not None:
                    # we get the index after calling `set_index` because this will give
                    # us the right index both when `zone_index` is a column name or a
                    # list-like.
                    # note that if zone_index is a list, pandas will try to interpret
                    # its values as column names, so we need to convert it to a numpy
                    # array/pandas series first so that the values are set as index.
                    # we will convert it to a pandas series so that we can set a name.
                    if isinstance(zone_index, list):
                        zone_index = pd.Series(zone_index, name="zone")
                    zone_index = zones.set_index(zone_index).index
                    # we now take just the "geometry" column and treat `zones` as
                    # GeoSeries.
                    zones = zones.geometry
                else:
                    # take just the "geometry" column, treat `zones` as GeoSeries but
                    # rename the index to "zone"
                    zones = zones.geometry.rename_axis("zone")

            # at this point, `zones` must be a geo-series or a list-like of shapely
            # geometries
            if not isinstance(zones, gpd.GeoSeries):
                # convert to a geo-series with the CRS of the landscape raster
                zones = gpd.GeoSeries(zones, crs=src.crs)
            else:
                # `zones` must be a geo-series, reproject into the CRS of the landscape
                # if needed
                if zones.crs != src.crs:
                    zones = zones.to_crs(src.crs)
            if zone_index is not None:
                # set the geo-series index
                zones = zones.set_axis(zone_index)

            # now that we have a processed geo series, store it as an instance attribute
            self.zone_gser = zones

            # now perform a masked read of the raster for the zone geometries
            landscapes = [
                Landscape(
                    zone_arr[0],
                    res=src.res,
                    nodata=src.nodata,
                    transform=zone_transform,
                    neighborhood_rule=neighborhood_rule,
                )
                for zone_arr, zone_transform in [
                    mask.mask(src, [geom], crop=True) for geom in zones
                ]
            ]

            # start: raster-based alternative ------------------------------------------
            # rasterize vector features into a labeled array and use it to generate the
            # zone landscapes
            # if types.is_list_like(zones):
            #     # if isinstance(zones[0], (geometry.Polygon, geometry.MultiPolygon)):
            #     zones = features.rasterize(
            #     [
            #         (geom, attribute_val)
            #         for attribute_val, geom in zip(
            #             attribute_values, zones.to_crs(src.crs)
            #         )
            #     ],
            #     out_shape=src.shape,
            #     transform=landscape_transform,
            #     fill=src.nodata,
            # )

            # # finally, zones is a labeled array
            # landscape_arr = src.read(1)
            # landscapes = [
            #     Landscape(landscape_arr[loc]) for loc in ndimage.find_objects(zones)
            # ]
            # end: raster-based alternative --------------------------------------------

        # now call the parent's init
        # for the parent class (MultiLandscape), set:
        # * `attribute_name` (by order of preference): (i) the series' index name if
        #    non-None, (ii) the series' name if non-None, or (iii) provide a default
        # * `attribute_values`: the index of the zones geo-series
        super().__init__(
            landscapes,
            self.zone_gser.index.name or self.zone_gser.name or "zone",
            self.zone_gser.index.values,
        )

    def compute_zonal_statistics_gdf(  # noqa: D102
        self,
        *,
        metrics=None,
        class_val=None,
        metrics_kwargs=None,
    ):
        if class_val is not None:
            zonal_metrics_df = self.compute_class_metrics_df(
                metrics=metrics, classes=[class_val], metrics_kwargs=metrics_kwargs
            ).loc[class_val]
        else:
            zonal_metrics_df = self.compute_landscape_metrics_df(
                metrics=metrics, metrics_kwargs=metrics_kwargs
            )

        # ensure that we have numeric types (not strings)
        # metric_ser = pd.to_numeric(metric_ser)

        # return a geo-data frame
        zone_name = self.zone_gser.index.name
        return gpd.GeoDataFrame(
            zonal_metrics_df.pivot_table(
                index=zone_name,
                columns=zonal_metrics_df.index.names.difference([zone_name]),
            ),
            geometry=self.zone_gser,
        )

    compute_zonal_statistics_gdf.__doc__ = _compute_zonal_statistics_gdf_doc.format(
        col_return="metrics"
    )


class BufferAnalysis(ZonalAnalysis):
    """Buffer analysis around a feature of interest."""

    def __init__(
        self,
        landscape_filepath,
        base_geom,
        buffer_dists,
        *,
        buffer_rings=False,
        base_geom_crs=None,
        neighborhood_rule=None,
    ):
        """Initialize the buffer analysis.

        Parameters
        ----------
        landscape_filepath : str, file-like object or pathlib.Path object
            A string/file-like object/pathlib.Path object with the landscape data.
        base_geom : shapely geometry, geopandas.GeoSeries or geopandas.GeoDataFrame
            Geometry that will serve as a base to buffer around.
        buffer_dists : list-like
            Buffer distances, in units of the landscape CRS.
        buffer_rings : bool, default Falsenn
            If `False`, each buffer zone will consist of the whole region that lies
            within the respective buffer distance around the base geometry. If `True`,
            buffer zones will take the form of rings around the base geometry.
        base_geom_crs : str, dict or pyproj.CRS, optional
            The coordinate reference system of the base geometry. Required if the base
            geometry is a shapely geometry or a geopandas GeoSeries without the `crs`
            attribute set.
        neighborhood_rule : {'8', '4'}, optional
            Neighborhood rule to determine patch adjacencies, i.e: '8' (queen's
            case/Moore neighborhood) or '4' (rook's case/Von Neumann neighborhood).
            Ignored if `landscape` is a `Landscape` instance. If no value is provided,
            the default value set in `settings.DEFAULT_NEIGHBORHOOD_RULE` will be taken.
        """
        # 1. get a GeoSeries with the base geometry
        if isinstance(base_geom, geometry_base.BaseGeometry):
            if base_geom_crs is None:
                raise ValueError(
                    "If `base_geom` is a shapely geometry, `base_geom_crs` must be"
                    " provided."
                )
            base_gser = gpd.GeoSeries(base_geom, crs=base_geom_crs)
        else:
            # we assume that `base_geom` is a geopandas GeoSeries/GeoDataFrame
            if isinstance(base_geom, gpd.GeoDataFrame):
                # in this case, we just select the geometry column and then treat it
                # like a geo series
                base_geom = base_geom.geometry
            if base_geom.crs is None:
                if base_geom_crs is None:
                    raise ValueError(
                        "If `base_geom` is a naive geopandas GeoSeries/GeoDataFrame"
                        " (with no crs set), `base_geom_crs` must be provided."
                    )

                base_gser = base_geom.copy()  # avoid alias/ref problems
                base_gser.crs = base_geom_crs
            else:
                base_gser = base_geom

        # 3. buffer around base mask
        avg_longitude = (
            base_gser.to_crs("+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs")
            .union_all()
            .centroid.x
        )
        # trick from OSMnx to be able to buffer in meters
        utm_zone = int(np.floor((avg_longitude + 180) / 6.0) + 1)
        # utm_crs = {
        #     'datum': 'WGS84',
        #     'ellps': 'WGS84',
        #     'proj': 'utm',
        #     'zone': utm_zone,
        #     'units': 'm'
        # }
        utm_crs = (
            f"+proj=utm +zone={utm_zone} +ellps=WGS84 +datum=WGS84 +units=m +no_defs"
        )
        base_proj_geom = base_gser.to_crs(utm_crs).iloc[0]
        if buffer_rings:
            if not isinstance(base_proj_geom, geometry.Point):
                raise ValueError(
                    "Buffer rings can only work when `base_geom` is a `Point`."
                )
            _buffer_dists = np.concatenate([[0], buffer_dists])
            buffer_dists = list(
                map(
                    lambda d: "{}-{}".format(d[0], d[1]),
                    zip(_buffer_dists[:-1], _buffer_dists[1:]),
                )
            )
            zone_gser = gpd.GeoSeries(
                [
                    base_proj_geom.buffer(_buffer_dists[i + 1])
                    - base_proj_geom.buffer(_buffer_dists[i])
                    for i in range(len(_buffer_dists) - 1)
                ],
                index=buffer_dists,
                crs=utm_crs,
            )
        else:
            zone_gser = gpd.GeoSeries(
                [base_proj_geom.buffer(buffer_dist) for buffer_dist in buffer_dists],
                index=buffer_dists,
                crs=utm_crs,
            )

        # now we can call the parent's init with the landscape and the constructed
        # buffer geoseries
        super().__init__(
            landscape_filepath,
            zones=zone_gser.rename_axis("buffer_dist"),  # set the index name
            neighborhood_rule=neighborhood_rule,
        )

    # override docs
    def compute_class_metrics_df(  # noqa: D102
        self, *, metrics=None, classes=None, metrics_kwargs=None, fillna=None
    ):
        return super().compute_class_metrics_df(
            metrics=metrics,
            classes=classes,
            metrics_kwargs=metrics_kwargs,
            fillna=fillna,
        )

    compute_class_metrics_df.__doc__ = (
        multilandscape._compute_class_metrics_df_doc.format(
            index_descr="multi-indexed by the class and buffer distance",
            index_return="class, buffer distance (multi-index)",
        )
    )

    def compute_landscape_metrics_df(  # noqa: D102
        self, *, metrics=None, metrics_kwargs=None
    ):
        return super().compute_landscape_metrics_df(
            metrics=metrics, metrics_kwargs=metrics_kwargs
        )

    compute_landscape_metrics_df.__doc__ = (
        multilandscape._compute_landscape_metrics_df_doc.format(
            index_descr="indexed by the buffer distance",
            index_return="buffer distance (index)",
        )
    )


class ZonalGridAnalysis(ZonalAnalysis):
    """Zonal analysis over a grid."""

    @staticmethod
    def _get_grid_gser(
        bounds, num_zone_rows, num_zone_cols, zone_width, zone_height, offset
    ):
        # get the zone dimensions to generate the grid
        left, bottom, right, top = bounds

        total_width = right - left
        total_height = top - bottom

        # make sure that we have both the number of zone rows/columns and the zone
        # width/height
        if zone_width is None:
            try:
                zone_width = np.ceil(total_width / num_zone_cols)
            except TypeError:
                # num_zone_cols is also None
                raise ValueError(
                    "Either `num_zone_cols` or `zone_width` must be provided"
                )
        if zone_height is None:
            try:
                zone_height = np.ceil(total_height / num_zone_rows)
            except TypeError:
                # num_zone_rows is also None
                raise ValueError(
                    "Either `num_zone_rows` or `zone_height` must be provided"
                )
        if num_zone_cols is None:
            num_zone_cols = int(np.ceil(total_width / zone_width))

        if num_zone_rows is None:
            num_zone_rows = int(np.ceil(total_height / zone_height))

        # once we have the number of zone rows/columns and the zone width/height, we can
        # compute the grid
        if offset == "center":
            # center the grid on the raster bounds
            left = left - (num_zone_cols * zone_width - total_width) / 2
            top = top + (num_zone_rows * zone_height - total_height) / 2

        # generate a grid of size using numpy meshgrid
        grid_x, grid_y = np.meshgrid(
            np.arange(num_zone_cols) * zone_width + left,
            top - np.arange(num_zone_rows) * zone_height,
            indexing="xy",
        )

        # vectorize the grid as a geo series
        flat_grid_x = grid_x.flatten()
        flat_grid_y = grid_y.flatten()
        zones = pd.DataFrame(
            {
                "xmin": flat_grid_x,
                "ymin": flat_grid_y - zone_height,
                "xmax": flat_grid_x + zone_width,
                "ymax": flat_grid_y,
            }
        ).apply(lambda row: geometry.box(*row), axis=1)

        # # identify zones as their (row, col) position
        # zone_ids = np.array(
        #     [(row, col)
        #      for row in range(num_zone_rows) for col in range(num_zone_cols)]
        # )
        # return grid_gser.set_index(zone_ids)

        return zones

    def __init__(
        self,
        landscape_filepath,
        *,
        num_zone_cols=None,
        num_zone_rows=None,
        zone_width=None,
        zone_height=None,
        offset=None,
        neighborhood_rule=None,
    ):
        """Initialize the zonal grid analysis.

        Parameters
        ----------
        landscape_filepath : str, file-like object or pathlib.Path object
            A string/file-like object/pathlib.Path object with the landscape data.
        num_zone_cols, num_zone_rows : int, optional
            The number of zone columns/rows into which the landscape will be separated.
            If the landscape dimensions and the desired zones do not divide evenly, the
            zones will be defined for the minimum superset that covers the landscape
            bounds. If not provided, then `num_width`/`num_height` must be provided.
        zone_width, zone_height : numeric, optional
            The width/height of each zone (in units of the landscape CRS). If the
            landscape dimensions and the desired zones do not divide evenly, the zones
            will be defined for the minimum superset that covers the landscape bounds.
            If not provided, then `num_width`/`num_height` must be provided.
        offset : str, optional
            If set to "center", the and the landscape dimensions and the desired zones
            do not divide evenly, the grid is offsetted so that the landscape bounds are
            in the center of the grid. Otherwise, the grid starts at the top-left corner
            of the landscape. Ignored if the landscape dimensions and the desired zones
            divide evenly.
        neighborhood_rule : {'8', '4'}, optional
            Neighborhood rule to determine patch adjacencies, i.e: '8' (queen's
            case/Moore neighborhood) or '4' (rook's case/Von Neumann neighborhood). If
            no value is provided, the default value set in
            `settings.DEFAULT_NEIGHBORHOOD_RULE` will be taken.
        """
        with rio.open(landscape_filepath) as src:
            zone_gser = gpd.GeoSeries(
                ZonalGridAnalysis._get_grid_gser(
                    src.bounds,
                    num_zone_rows,
                    num_zone_cols,
                    zone_width,
                    zone_height,
                    offset,
                ),
                crs=src.crs,
            )

            # filter out zones that do not meet any valid data pixel
            def has_valid_data(geom):
                zone_arr, _ = mask.mask(src, [geom], crop=True)
                return np.any(zone_arr != src.nodata)

            zone_gser = zone_gser[zone_gser.apply(has_valid_data)]

        # now we can call the parent's init with the landscape and the constructed grid
        # geoseries
        super().__init__(
            landscape_filepath,
            zones=zone_gser.rename_axis("grid_cell"),  # set the index name
            neighborhood_rule=neighborhood_rule,
        )

    def plot_landscapes(self, *, cmap=None, ax=None, figsize=None, **plot_kwargs):
        """Plot the spatial distribution of the landscape zones.

        Parameters
        ----------
        cmap : str or `~matplotlib.colors.Colormap`, optional
            A Colormap instance.
        ax : axis object, optional
            Plot in given axis; if None creates a new figure.
        figsize : tuple of two numeric types, optional
            Size of the figure to create. Ignored if axis `ax` is provided.
        **plot_kwargs : optional
            Keyword arguments to be passed to `geopandas.GeoSeries.plot`.

        Returns
        -------
        ax : matplotlib.axes.Axes
            Returns the `Axes` object with the plot drawn onto it.
        """
        if cmap is None:
            cmap = plt.rcParams["image.cmap"]

        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
            ax.set_aspect("equal")

        if plot_kwargs is None:
            plot_kwargs = {}

        gpd.GeoDataFrame(
            {"color": np.arange(len(self.zone_gser))}, geometry=self.zone_gser
        ).plot("color", ax=ax, cmap=cmap, **plot_kwargs)

        return ax



"""Build extensions."""

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

extensions = [
    Extension(
        name="pylandstats.adjacency",
        sources=["pylandstats/adjacency.pyx"],
        include_dirs=[np.get_include()],
    )
]

setup(
    ext_modules=cythonize(extensions),
)



"""pylandstats tests."""

import os
import shutil
import unittest
import warnings
from os import path

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio as rio
from pandas.testing import assert_index_equal
from shapely import geometry
from sklearn import decomposition, impute, preprocessing

import pylandstats as pls

plt.switch_backend("agg")  # only for testing purposes
geom_crs = "epsg:4326"

tests_dir = "tests"
tests_data_dir = path.join(tests_dir, "data")


class TestImports(unittest.TestCase):
    def test_base_imports(self):
        pass

    def test_geo_imports(self):
        pass


class TestLandscape(unittest.TestCase):
    def setUp(self):
        self.ls_arr = np.load(
            path.join(tests_data_dir, "ls250_06.npy"), allow_pickle=True
        )
        self.ls = pls.Landscape(self.ls_arr, res=(250, 250))
        self.landscape_fp = path.join(tests_data_dir, "ls250_06.tif")

    def test_io(self):
        # test that if we provide a ndarray, we also need to provide the resolution
        with self.assertRaises(ValueError) as cm:
            ls = pls.Landscape(self.ls_arr)
        self.assertIn("must be provided", str(cm.exception))

        ls = pls.Landscape(self.landscape_fp)
        # resolutions are not exactly 250, they are between [249, 251], so we need to
        # use a large delta
        self.assertAlmostEqual(ls.cell_width, 250, delta=1)
        self.assertAlmostEqual(ls.cell_height, 250, delta=1)
        self.assertAlmostEqual(ls.cell_area, 250 * 250, delta=250)

        # test that the transform is None if we instantiate a landscape from an ndarray
        # (without providing the `transform` argument, but that it is not none we get
        # the landscape transform from a raster path
        self.assertIsNone(self.ls.transform)
        self.assertIsNotNone(ls.transform)

    def test_metrics_parameters(self):
        ls = self.ls

        for patch_metric in pls.Landscape.PATCH_METRICS:
            method = getattr(ls, patch_metric)
            self.assertIsInstance(method(), pd.DataFrame)
            self.assertIsInstance(method(class_val=ls.classes[0]), pd.Series)

        for class_metric in pls.Landscape.CLASS_METRICS:
            self.assertTrue(
                np.isreal(getattr(ls, class_metric)(class_val=ls.classes[0]))
            )

        for landscape_metric in pls.Landscape.LANDSCAPE_METRICS:
            self.assertTrue(np.isreal(getattr(ls, landscape_metric)()))

    def test_neighborhood_rules(self):
        # test that providing a value different than 8 or 4 raises a `ValueError`
        with self.assertRaises(ValueError) as cm:
            pls.Landscape(self.landscape_fp, neighborhood_rule="2")
        self.assertIn("is not among", str(cm.exception))
        # test that we can provide the argument as int as long as it is 8 or 4
        for neighborhood_rule in (8, 4):
            self.assertEqual(
                pls.Landscape(
                    self.landscape_fp, neighborhood_rule=neighborhood_rule
                ).neighborhood_rule,
                str(neighborhood_rule),
            )
        # test that there is at least the same number of patches with the Moore
        # neighborhood than with Von Neumann's
        ls_moore = pls.Landscape(self.landscape_fp, neighborhood_rule="8")
        ls_von_neumann = pls.Landscape(self.landscape_fp, neighborhood_rule="4")
        self.assertLessEqual(
            ls_moore.number_of_patches(), ls_von_neumann.number_of_patches()
        )

    def test_metrics_warnings(self):
        # test that warnings are raised

        # class-level metrics
        # euclidean nearest neighbor will return nan (and raise an informative warning)
        # if there is not at least two patches of each class. Let us test this by
        # creating a landscape with a background of class 1 and a single patch of class
        # 2
        arr = np.ones((4, 4))
        arr[1:-1, 1:-1] = 2
        ls = pls.Landscape(arr, res=(1, 1))

        # let us test that both the computation at the class-level (1 and 2) and at the
        # landscape level (`class_val` of `None`) raise at least one warning (the exact
        # number of warnings raised can be different in Python 2 and 3)
        for class_val in [1, 2, None]:
            with warnings.catch_warnings(record=True) as w:
                ls.euclidean_nearest_neighbor(class_val=class_val)
                self.assertGreater(len(w), 0)

        # landscape-level metrics
        # some landscape-level metrics require at least two classes.
        ls = pls.Landscape(np.ones((4, 4)), res=(1, 1))
        for method in pls.Landscape.ENTROPY_METRICS:
            with warnings.catch_warnings(record=True) as w:
                getattr(ls, method)()
                self.assertGreater(len(w), 0)

    def test_metric_dataframes(self):
        ls = self.ls
        patch_metrics = set(pls.Landscape.PATCH_METRICS)
        class_metrics = set(pls.Landscape.CLASS_METRICS)
        landscape_metrics = set(pls.Landscape.LANDSCAPE_METRICS)

        patch_df = ls.compute_patch_metrics_df()
        self.assertTrue(
            np.all(patch_df.columns.drop("class_val") == pls.Landscape.PATCH_METRICS)
        )
        self.assertEqual(patch_df.index.name, "patch_id")
        # try that raised ValueErrors have different error messages depending on the
        # context
        with self.assertRaises(ValueError) as cm:
            ls.compute_patch_metrics_df(metrics=["foo"])
        self.assertIn("is not among", str(cm.exception))
        for metric in class_metrics.union(landscape_metrics):
            with self.assertRaises(ValueError) as cm:
                ls.compute_patch_metrics_df(metrics=[metric])
            self.assertIn("cannot be computed", str(cm.exception))

        class_df = ls.compute_class_metrics_df()
        self.assertEqual(
            len(class_df.columns.difference(pls.Landscape.CLASS_METRICS)), 0
        )
        self.assertEqual(class_df.index.name, "class_val")
        # try that raised ValueErrors have different error messages depending on the
        # context
        with self.assertRaises(ValueError) as cm:
            ls.compute_class_metrics_df(metrics=["foo"])
        self.assertIn("is not among", str(cm.exception))
        for metric in patch_metrics.union(landscape_metrics.difference(class_metrics)):
            with self.assertRaises(ValueError) as cm:
                ls.compute_class_metrics_df(metrics=[metric])
            self.assertIn("cannot be computed", str(cm.exception))

        landscape_df = ls.compute_landscape_metrics_df()
        self.assertEqual(
            len(landscape_df.columns.difference(pls.Landscape.LANDSCAPE_METRICS)),
            0,
        )
        self.assertEqual(len(landscape_df.index), 1)
        # try that raised ValueErrors have different error messages depending on the
        # context
        with self.assertRaises(ValueError) as cm:
            ls.compute_landscape_metrics_df(metrics=["foo"])
        self.assertIn("is not among", str(cm.exception))
        for metric in patch_metrics.union(class_metrics.difference(landscape_metrics)):
            with self.assertRaises(ValueError) as cm:
                ls.compute_landscape_metrics_df(metrics=[metric])
            self.assertIn("cannot be computed", str(cm.exception))

    def test_metrics_value_ranges(self):
        ls = self.ls

        # basic tests of the `Landscape` class' attributes
        self.assertNotIn(ls.nodata, ls.classes)
        self.assertGreater(ls.landscape_area, 0)

        class_val = ls.classes[0]
        # label_arr = ls._get_label_arr(class_val)

        # patch-level metrics
        self.assertTrue((ls.area()["area"] > 0).all())
        self.assertTrue((ls.perimeter()["perimeter"] > 0).all())
        self.assertTrue((ls.perimeter_area_ratio()["perimeter_area_ratio"] > 0).all())
        self.assertTrue((ls.shape_index()["shape_index"] >= 1).all())
        _fractal_dimension_ser = ls.fractal_dimension()["fractal_dimension"]
        self.assertTrue((_fractal_dimension_ser <= 2).all())
        # ACHTUNG: ugly hardcoded tolerance to correct for mysterious errors in GitHub
        # Actions with some Python versions
        self.assertTrue((_fractal_dimension_ser >= 1 - 1e-3).all())
        self.assertTrue((ls.core_area()["core_area"] >= 0).all())
        self.assertTrue((ls.number_of_core_areas()["number_of_core_areas"] >= 0).all())
        _core_area_index_ser = ls.core_area_index()["core_area_index"]
        self.assertTrue((_core_area_index_ser >= 0).all())
        self.assertTrue((_core_area_index_ser < 100).all())
        # TODO: assert 0 <= ls.contiguity_index(patch_arr) <= 1
        # ACHTUNG: euclidean nearest neighbor can be nan for classes with less than two
        # patches
        self.assertTrue(
            (
                ls.euclidean_nearest_neighbor()["euclidean_nearest_neighbor"].dropna()
                > 0
            ).all()
        )
        # TODO: assert 0 <= ls.proximity(patch_arr) <= 1

        # class-level metrics
        self.assertGreater(ls.total_area(class_val=class_val), 0)
        self.assertTrue(0 < ls.proportion_of_landscape(class_val) < 100)
        self.assertTrue(ls.patch_density(class_val=class_val) > 0)
        self.assertTrue(0 < ls.largest_patch_index(class_val=class_val) < 100)
        self.assertGreaterEqual(ls.total_edge(class_val=class_val), 0)
        self.assertGreaterEqual(ls.edge_density(class_val=class_val), 0)
        self.assertGreaterEqual(ls.total_core_area(class_val=class_val), 0)
        self.assertGreaterEqual(
            ls.core_area_proportion_of_landscape(class_val=class_val), 0
        )
        self.assertGreaterEqual(
            ls.number_of_disjunct_core_areas(class_val=class_val), 0
        )

        # the value ranges of mean, area-weighted mean and median aggregations are going
        # to be the same as their respective original metrics
        mean_suffixes = ["_mn", "_am", "_md"]
        # the value ranges of the range, standard deviation and coefficient of variation
        # will always be nonnegative as long as the means are nonnegative as well (which
        # is the case of all of the metrics implemented so far)
        var_suffixes = ["_ra", "_sd", "_cv"]

        for mean_suffix in mean_suffixes:
            self.assertGreater(
                getattr(ls, "area" + mean_suffix)(class_val=class_val), 0
            )
            self.assertGreater(
                getattr(ls, "perimeter_area_ratio" + mean_suffix)(class_val=class_val),
                0,
            )
            self.assertGreaterEqual(
                getattr(ls, "shape_index" + mean_suffix)(class_val=class_val), 1
            )
            self.assertTrue(
                1
                <= getattr(ls, "fractal_dimension" + mean_suffix)(class_val=class_val)
                <= 2
            )
            self.assertGreaterEqual(
                getattr(ls, "core_area" + mean_suffix)(class_val=class_val), 0
            )
            self.assertGreaterEqual(
                getattr(ls, "disjunct_core_area" + mean_suffix)(class_val=class_val), 0
            )
            # self.assertGreaterEqual(
            #     getattr(ls, "number_of_core_areas" + mean_suffix)(
            #         class_val=class_val),
            #     0,
            # )
            cai = getattr(ls, "core_area_index" + mean_suffix)(class_val=class_val)
            self.assertGreaterEqual(cai, 0)
            self.assertLess(cai, 100)
            # assert 0 <= getattr(
            #     ls, 'contiguity_index' + mean_suffix)(class_val) <= 1
            # assert getattr(ls, 'proximity' + mean_suffix)(class_val) >= 0
            # ACHTUNG: euclidean nearest neighbor can be nan for classes with less than
            # two patches
            enn = getattr(ls, "euclidean_nearest_neighbor" + mean_suffix)(
                class_val=class_val
            )
            self.assertTrue(enn > 0 or np.isnan(enn))

        for var_suffix in var_suffixes:
            self.assertGreaterEqual(
                getattr(ls, "area" + mean_suffix)(class_val=class_val), 0
            )
            self.assertGreaterEqual(
                getattr(ls, "perimeter_area_ratio" + var_suffix)(class_val=class_val), 0
            )
            self.assertGreaterEqual(
                getattr(ls, "shape_index" + var_suffix)(class_val=class_val), 0
            )
            self.assertGreaterEqual(
                getattr(ls, "fractal_dimension" + var_suffix)(class_val=class_val), 0
            )
            # assert getattr(
            #    ls, 'contiguity_index' + var_suffix)(class_val) >= 0
            # assert getattr(ls, 'proximity' + var_suffix)(class_val) >= 0
            # ACHTUNG: euclidean nearest neighbor can be nan for classes with less than
            # two patches
            enn = getattr(ls, "euclidean_nearest_neighbor" + var_suffix)(
                class_val=class_val
            )
            self.assertTrue(enn >= 0 or np.isnan(enn))

        # TODO: assert 0 < ls.interspersion_juxtaposition_index(
        #           class_val) <= 100
        self.assertGreaterEqual(ls.landscape_shape_index(class_val=class_val), 1)
        self.assertTrue(
            ls.cell_area / ls.landscape_area
            <= ls.effective_mesh_size(class_val=class_val)
            <= ls.landscape_area,
            1,
        )

        # landscape-level metrics
        self.assertGreater(ls.total_area(), 0)
        self.assertGreater(ls.patch_density(), 0)
        self.assertTrue(0 < ls.largest_patch_index() < 100)
        self.assertGreaterEqual(ls.total_edge(), 0)
        self.assertGreaterEqual(ls.edge_density(), 0)
        self.assertTrue(0 < ls.largest_patch_index() <= 100)
        self.assertGreaterEqual(ls.total_edge(), 0)
        self.assertGreaterEqual(ls.edge_density(), 0)

        # for class_val in ls.classes:
        #     print('num_patches', class_val, ls._get_num_patches(class_val))
        #     print('patch_areas', len(ls._get_patch_areas(class_val)))

        # raise ValueError

        for mean_suffix in mean_suffixes:
            self.assertGreater(getattr(ls, "area" + mean_suffix)(), 0)
            self.assertGreater(getattr(ls, "perimeter_area_ratio" + mean_suffix)(), 0)
            self.assertGreaterEqual(getattr(ls, "shape_index" + mean_suffix)(), 1)
            self.assertTrue(1 <= getattr(ls, "fractal_dimension" + mean_suffix)() <= 2)
            # assert 0 <= getattr(ls, 'contiguity_index' + mean_suffix)() <= 1
            # assert getattr(ls, 'proximity' + mean_suffix)() >= 0
            # ACHTUNG: euclidean nearest neighbor can be nan for classes with less than
            # two patches
            enn = getattr(ls, "euclidean_nearest_neighbor" + mean_suffix)()
            self.assertTrue(enn > 0 or np.isnan(enn))
        for var_suffix in var_suffixes:
            self.assertGreater(getattr(ls, "area" + var_suffix)(), 0)
            self.assertGreaterEqual(
                getattr(ls, "perimeter_area_ratio" + var_suffix)(), 0
            )
            self.assertGreaterEqual(getattr(ls, "shape_index" + var_suffix)(), 0)
            self.assertGreaterEqual(getattr(ls, "fractal_dimension" + var_suffix)(), 0)
            # assert getattr(ls, 'contiguity_index' + var_suffix)() >= 0
            # assert getattr(ls, 'proximity' + var_suffix)() >= 0
            # ACHTUNG: euclidean nearest neighbor can be nan for classes with less than
            # two patches
            enn = getattr(ls, "euclidean_nearest_neighbor" + var_suffix)()
            self.assertTrue(enn >= 0 or np.isnan(enn))

        self.assertGreaterEqual(ls.landscape_shape_index(), 1)
        self.assertTrue(
            ls.cell_area / ls.landscape_area
            <= ls.effective_mesh_size()
            <= ls.landscape_area,
            1,
        )
        for method in [
            "entropy",
            "shannon_diversity_index",
            "joint_entropy",
            "conditional_entropy",
            "mutual_information",
        ]:
            self.assertGreaterEqual(getattr(ls, method)(), 0)
        self.assertTrue(0 < ls.relative_mutual_information(), 1)
        self.assertTrue(0 < ls.contagion() <= 100)

        # TODO: assert 0 < ls.interspersion_juxtaposition_index() <= 100

    def test_transonic(self):
        ls_arr = np.load(path.join(tests_data_dir, "ls250_06.npy"), allow_pickle=True)
        ls = pls.Landscape(ls_arr, res=(250, 250))
        adjacency_df = ls._adjacency_df
        self.assertIsInstance(adjacency_df, pd.DataFrame)

    def test_plot_landscape(self):
        # first test for a landscape without affine transform (instantiated from an
        # ndarray and without providing a non-None `transform` argument)
        ax = self.ls.plot_landscape()
        # returned axis must be instances of matplotlib axes
        self.assertIsInstance(ax, plt.Axes)

        # now do the same test for a landscape with affine transform (e.g., instantiated
        # from a raster file)
        ls = pls.Landscape(self.landscape_fp)
        ax = ls.plot_landscape()
        self.assertIsInstance(ax, plt.Axes)
        # and further test that the plot bounds correspond to the transform's offsets
        self.assertAlmostEqual(ax.get_xlim()[0], ls.transform.xoff)
        self.assertAlmostEqual(ax.get_ylim()[1], ls.transform.yoff)

        # test legend arguments
        self.assertIsNone(ls.plot_landscape(legend=False).get_legend())
        self.assertIsNotNone(ls.plot_landscape(legend=True))
        self.assertIsNotNone(
            ls.plot_landscape(legend=True, legend_kwargs=dict(loc="center right"))
        )


class TestMultiLandscape(unittest.TestCase):
    def setUp(self):
        from pylandstats import multilandscape

        self.landscapes = [
            pls.Landscape(
                np.load(path.join(tests_data_dir, "ls100_06.npy"), allow_pickle=True),
                res=(100, 100),
            ),
            pls.Landscape(
                np.load(path.join(tests_data_dir, "ls250_06.npy"), allow_pickle=True),
                res=(250, 250),
            ),
        ]
        self.landscape_fps = [
            path.join(tests_data_dir, "ls100_06.tif"),
            path.join(tests_data_dir, "ls250_06.tif"),
        ]
        self.attribute_name = "resolution"
        self.attribute_values = [100, 250]
        self.inexistent_class_val = 999

        # use this class just for testing purposes
        class InstantiableMultiLandscape(multilandscape.MultiLandscape):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

        self.InstantiableMultiLandscape = InstantiableMultiLandscape

    def test_multilandscape_init(self):
        from pylandstats import multilandscape

        # test that we cannot instantiate an abstract class
        self.assertRaises(TypeError, multilandscape.MultiLandscape)

        # test that if we init a MultiLandscape from filepaths, Landscape instances are
        # automatically built
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps, self.attribute_name, self.attribute_values
        )
        for landscape in ml.landscape_ser:
            self.assertIsInstance(landscape, pls.Landscape)
        # test that we can pass keyword arguments to the `Landscape` instantiation when
        # providing filepaths
        landscape_kwargs = {"neighborhood_rule": "4"}
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps,
            self.attribute_name,
            self.attribute_values,
            **landscape_kwargs,
        )
        for landscape in ml.landscape_ser:
            self.assertEqual(
                landscape.neighborhood_rule, landscape_kwargs["neighborhood_rule"]
            )
        # test that if we instantiate providing `Landscape` instances, the provided
        # keyword arguments are ignored
        ml = self.InstantiableMultiLandscape(
            self.landscapes,
            self.attribute_name,
            self.attribute_values,
            **landscape_kwargs,
        )
        for landscape in ml.landscape_ser:
            self.assertNotEqual(
                landscape.neighborhood_rule, landscape_kwargs["neighborhood_rule"]
            )

        # from this point on, always instantiate from filepaths

        # test that constructing a MultiLandscape where the list of values of the
        # identifying attributes `attribute_values` (in this example, the list of
        # resolutions `[100, 250]`) mismatches the length of the list of landscapes
        # raises a ValueError
        self.assertRaises(
            ValueError,
            self.InstantiableMultiLandscape,
            self.landscape_fps,
            self.attribute_name,
            [250],
        )

    def test_multilandscape_dataframes(self):
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps, self.attribute_name, self.attribute_values
        )
        # test that the data frames that result from `compute_class_metrics_df` and
        # `compute_landscape_metrics_df` are well constructed
        class_metrics_df = ml.compute_class_metrics_df()
        attribute_values = ml.landscape_ser.index
        self.assertTrue(np.all(class_metrics_df.columns == pls.Landscape.CLASS_METRICS))
        self.assertTrue(
            np.all(
                class_metrics_df.index
                == pd.MultiIndex.from_product([ml.present_classes, attribute_values])
            )
        )
        landscape_metrics_df = ml.compute_landscape_metrics_df()
        self.assertTrue(
            np.all(landscape_metrics_df.columns == pls.Landscape.LANDSCAPE_METRICS)
        )
        self.assertTrue(np.all(landscape_metrics_df.index == attribute_values))

        # now test the same but with an analysis that only considers a subset of metrics
        # and a subset of classes
        metrics = ["total_area", "edge_density", "proportion_of_landscape"]
        classes = ml.landscape_ser.iloc[0].classes[:2]

        class_metrics_df = ml.compute_class_metrics_df(metrics=metrics, classes=classes)
        attribute_values = ml.landscape_ser.index
        self.assertTrue(np.all(class_metrics_df.columns == metrics))
        self.assertTrue(
            np.all(
                class_metrics_df.index
                == pd.MultiIndex.from_product([classes, attribute_values])
            )
        )
        # 'proportion_of_landscape' cannot be computed at the landscape level
        # (TODO: test for that elsewhere)
        landscape_metrics = metrics[:2]
        landscape_metrics_df = ml.compute_landscape_metrics_df(
            metrics=landscape_metrics
        )
        self.assertTrue(np.all(landscape_metrics_df.columns == landscape_metrics))
        self.assertTrue(np.all(landscape_metrics_df.index == attribute_values))

    def test_multilandscape_class_metrics_fillna(self):
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps, self.attribute_name, self.attribute_values
        )
        fillna_df = ml.compute_class_metrics_df()
        nofillna_df = ml.compute_class_metrics_df(fillna=False)

        for metric in fillna_df.columns:
            if metric in self.InstantiableMultiLandscape.METRIC_FILLNA_DICT:
                # this metric should be zero in the first data frame and `NaN` in the
                # second one
                self.assertEqual(
                    np.count_nonzero(fillna_df[metric][nofillna_df[metric].isna()]),
                    0,
                )
            else:
                # the columns for this metric should be identical in both data frames
                self.assertTrue(fillna_df[metric].equals(nofillna_df[metric]))

    def test_multilandscape_metrics_kwargs(self):
        # Instantiate two multilandscape analyses, one with FRAGSTATS' defaults and the
        # other with keyword arguments specifying the total area in meters and including
        # the boundary in the computation of the total edge.
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps, self.attribute_name, self.attribute_values
        )
        # entropy metrics that accept a `base` keyword argument (all entropy metrics
        # except for Shannon's diversity index and contagion)
        entropy_metrics = [
            "entropy",
            "joint_entropy",
            "conditional_entropy",
            "mutual_information",
        ]
        metrics_kwargs = {
            "total_area": {"hectares": False},
            "perimeter_area_ratio_mn": {"hectares": True},
            "total_edge": {"count_boundary": True},
            **{entropy_metric: {"base": 10} for entropy_metric in entropy_metrics},
        }

        class_metrics_df = ml.compute_class_metrics_df()
        class_metrics_kwargs_df = ml.compute_class_metrics_df(
            metrics_kwargs=metrics_kwargs
        )
        landscape_metrics_df = ml.compute_landscape_metrics_df()
        landscape_metrics_kwargs_df = ml.compute_landscape_metrics_df(
            metrics_kwargs=metrics_kwargs
        )
        for attribute_value in ml.landscape_ser.index:
            # For all attribute values and all classes, metric values in hectares should
            # be less than in meters, and excluding boundaries should be less or equal
            # than including them
            landscape_metrics = landscape_metrics_df.loc[attribute_value]
            landscape_metrics_kwargs = landscape_metrics_kwargs_df.loc[attribute_value]
            self.assertLess(
                landscape_metrics["total_area"],
                landscape_metrics_kwargs["total_area"],
            )
            self.assertLessEqual(
                landscape_metrics["total_edge"],
                landscape_metrics_kwargs["total_edge"],
            )

            for class_val in ml.present_classes:
                class_metrics = class_metrics_df.loc[class_val, attribute_value]
                class_metrics_kwargs = class_metrics_kwargs_df.loc[
                    class_val, attribute_value
                ]

                # It could be that for some attribute values, some classes are not
                # present within the respective Landscape. If so, some metrics (e.g.,
                # 'perimeter_area_ratio_mn') will be `nan`, both for the analysis with
                # and without keyword arguments.
                # For area and edge metrics, PyLandStats interprets such `nan` values as
                # 0, which is why we need to include the equality in the comparison
                if class_metrics.isnull().all():
                    self.assertTrue(class_metrics_kwargs.isnull().all())
                else:
                    self.assertLessEqual(
                        class_metrics["total_area"],
                        class_metrics_kwargs["total_area"],
                    )
                    # we need quite some tolerance because pixel resolutions in raster
                    # files might be weird float values, e.g., 99.13213 instead of 100
                    # (meters)
                    self.assertLessEqual(
                        class_metrics["total_edge"],
                        1.01 * class_metrics_kwargs["total_edge"],
                    )

            # For all attribute values, entropy metric values computed with the default
            # base=2 should be greater than those computed with a custom base=10
            for entropy_metric in entropy_metrics:
                self.assertGreater(
                    landscape_metrics[entropy_metric],
                    landscape_metrics_kwargs[entropy_metric],
                )

    def test_multilandscape_plot_metrics(self):
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps, self.attribute_name, self.attribute_values
        )

        existent_class_val = ml.present_classes[0]
        # TODO: test legend and figsize

        # test that there is only one line when plotting a single metric at the
        # landscape level
        ax = ml.plot_metric("patch_density", class_val=None)
        self.assertEqual(len(ax.lines), 1)
        # test that there are two lines if we add the plot of a single metric (e.g., at
        # the level of an existent class) to the previous axis
        ax = ml.plot_metric("patch_density", class_val=existent_class_val, ax=ax)
        self.assertEqual(len(ax.lines), 2)
        # test that the x data of any line corresponds to the attribute values
        for line in ax.lines:
            self.assertTrue(np.all(line.get_xdata() == ml.landscape_ser.index))

        # test metric label arguments/settings
        # when passing default arguments, the axis ylabel must be the one set within the
        # settings module
        self.assertEqual(
            ml.plot_metric("edge_density").get_ylabel(),
            pls.settings.metric_label_dict["edge_density"],
        )
        # when passing `metric_legend=False`, the axis ylabel must be an empty string
        self.assertEqual(
            ml.plot_metric("edge_density", metric_legend=False).get_ylabel(),
            "",
        )
        # when passing an non-empty string as `metric_label`, the axis ylabel must be
        # such string
        self.assertEqual(
            ml.plot_metric("edge_density", metric_label="foo").get_ylabel(),
            "foo",
        )

        # test that asking to plot an inexistent metric raises a `ValueError` try that
        # raised ValueErrors have different error messages depending on the context
        with self.assertRaises(ValueError) as cm:
            ml.plot_metric("foo")
        self.assertIn("is not among", str(cm.exception))
        with self.assertRaises(ValueError) as cm:
            ml.plot_metric("proportion_of_landscape")
        self.assertIn("cannot be computed", str(cm.exception))
        with self.assertRaises(ValueError) as cm:
            ml.plot_metric("contagion", class_val=1)
        self.assertIn("cannot be computed", str(cm.exception))

    def test_plot_landscapes(self):
        ml = self.InstantiableMultiLandscape(
            self.landscape_fps, self.attribute_name, self.attribute_values
        )

        fig = ml.plot_landscapes()

        # there must be one column for each landscape
        self.assertEqual(len(fig.axes), len(ml))

        # returned axes must be instances of matplotlib axes
        for ax in fig.axes:
            self.assertIsInstance(ax, plt.Axes)

        # test that by default, the dimensions of the resulting will come from
        # matplotlib's settings
        rc_figwidth, rc_figheight = plt.rcParams["figure.figsize"]
        figwidth, figheight = fig.get_size_inches()
        # we have `len(ml)` axis, therefore the actual `figwidth` must be
        # `len(ml) * rc_figwidth`
        self.assertAlmostEqual(figwidth, len(ml) * rc_figwidth)
        self.assertAlmostEqual(figheight, rc_figheight)
        # if instead, we customize the figure size, the dimensions of the resulting
        # figure must be the customized ones
        custom_figsize = (10, 10)
        fig = ml.plot_landscapes(subplots_kwargs={"figsize": custom_figsize})
        figwidth, figheight = fig.get_size_inches()
        self.assertAlmostEqual(custom_figsize[0], figwidth)
        self.assertAlmostEqual(custom_figsize[1], figheight)


class TestSpatioTemporalAnalysis(unittest.TestCase):
    def setUp(self):
        # we will only test reading from filepaths because the consistency between
        # passing `Landscape` objects or filepaths is already tested in
        # `TestMultiLandscape`
        self.landscape_fps = [
            path.join(tests_data_dir, "ls250_06.tif"),
            path.join(tests_data_dir, "ls250_12.tif"),
        ]
        self.dates = [2006, 2012]
        self.inexistent_class_val = 999

    def test_spatiotemporalanalysis_init(self):
        # test that the landscape index name is dates, and that if the `dates` argument
        # is not provided when instantiating a `SpatioTemporalAnalysis`, the dates
        # attribute is properly and automatically generated
        sta = pls.SpatioTemporalAnalysis(self.landscape_fps)
        self.assertEqual(sta.landscape_ser.index.name, "dates")
        self.assertEqual(len(sta), len(sta.landscape_ser))

        # test the `neighborhood_rule` argument
        neighborhood_rule = "4"
        other_neighborhood_rule = "8"
        # test that if provided and the elements of `landscapes` are filepaths, the
        # value is passed to each landscape
        for ls in pls.SpatioTemporalAnalysis(
            self.landscape_fps,
            dates=self.dates,
            neighborhood_rule=neighborhood_rule,
        ).landscape_ser:
            self.assertEqual(ls.neighborhood_rule, neighborhood_rule)
        # test that if provided and the elements of `landscapes` are `Landscape`
        # instances, the value is ignored
        for ls in pls.SpatioTemporalAnalysis(
            [
                pls.Landscape(landscape_fp, neighborhood_rule=other_neighborhood_rule)
                for landscape_fp in self.landscape_fps
            ],
            dates=self.dates,
            neighborhood_rule=neighborhood_rule,
        ).landscape_ser:
            self.assertEqual(ls.neighborhood_rule, other_neighborhood_rule)
        # test that if not provided and the elements of `landscapes` are filepaths, the
        # default value is taken
        for ls in pls.SpatioTemporalAnalysis(
            self.landscape_fps, dates=self.dates
        ).landscape_ser:
            self.assertEqual(
                ls.neighborhood_rule, pls.settings.DEFAULT_NEIGHBORHOOD_RULE
            )

    def test_spatiotemporalanalysis_dataframes(self):
        # test with the default constructor
        sta = pls.SpatioTemporalAnalysis(self.landscape_fps)

        # test that the data frames that result from `compute_class_metrics_df` and
        # `compute_landscape_metrics_df` are well constructed
        class_metrics_df = sta.compute_class_metrics_df()
        self.assertTrue(
            np.all(
                class_metrics_df.index
                == pd.MultiIndex.from_product(
                    [sta.present_classes, sta.landscape_ser.index]
                )
            )
        )
        landscape_metrics_df = sta.compute_landscape_metrics_df()
        self.assertTrue(np.all(landscape_metrics_df.index == sta.landscape_ser.index))

        # now test the same but with an analysis that only considers a subset of metrics
        # and a subset of classes
        metrics = ["total_area", "edge_density", "proportion_of_landscape"]
        classes = sta.present_classes[:2]

        class_metrics_df = sta.compute_class_metrics_df(
            metrics=metrics, classes=classes
        )
        self.assertTrue(
            np.all(
                class_metrics_df.index
                == pd.MultiIndex.from_product([classes, sta.landscape_ser.index])
            )
        )
        # 'proportion_of_landscape' cannot be computed at the landscape level (TODO:
        # test for that elsewhere)
        landscape_metrics = metrics[:2]
        landscape_metrics_df = sta.compute_landscape_metrics_df(
            metrics=landscape_metrics
        )
        self.assertTrue(np.all(landscape_metrics_df.index == sta.landscape_ser.index))

    def test_spatiotemporalanalysis_plot_metrics(self):
        sta = pls.SpatioTemporalAnalysis(self.landscape_fps, dates=self.dates)

        # test for `None` (landscape-level) and an existing class (class-level)
        for class_val in [None, sta.present_classes[0]]:
            # test that the x data of the line corresponds to the dates
            self.assertTrue(
                np.all(
                    sta.plot_metric("patch_density", class_val=class_val)
                    .lines[0]
                    .get_xdata()
                    == self.dates
                )
            )


class TestZonaAlnalysis(unittest.TestCase):
    def setUp(self):
        self.landscape_fp = path.join(tests_data_dir, "ls250_06.tif")
        with rio.open(self.landscape_fp) as src:
            self.landscape_transform = src.transform
            self.landscape_crs = src.crs
        self.zones_fp = path.join(tests_data_dir, "gmb-lausanne.gpkg")
        self.zone_gdf = gpd.read_file(self.zones_fp)
        self.label_arr = np.load(
            path.join(tests_data_dir, "label_arr.npy"), allow_pickle=True
        )
        # for buffer analysis
        self.geom = geometry.Point(6.6327025, 46.5218269)
        self.buffer_dists = [10000, 15000, 20000]

        self.tmp_dir = path.join(tests_dir, "tmp")
        os.mkdir(self.tmp_dir)

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_zonal_init(self):
        # test passing GeoSeries, list-like of geometries, GeoDataFrame and geopandas
        # files as `zones`
        # first test the GeoSeries and list-like of shapely geometries, which work like
        # the others except that we cannot set a column as the zone index
        zone_gser = self.zone_gdf["geometry"].copy()
        za = pls.ZonalAnalysis(self.landscape_fp, zone_gser)
        self.assertLessEqual(len(za), len(self.zone_gdf))
        self.assertTrue(np.all(za.zone_gser.index == zone_gser.index))
        # also test that landscape index name is properly set when using geoseries as
        # `zones` first test for a geoseries with name and unnamed index
        zone_gser.name = "bar"
        za = pls.ZonalAnalysis(self.landscape_fp, zone_gser)
        self.assertEqual(za.landscape_ser.index.name, zone_gser.name)
        # now test that for a named geoseries with a named index, the geoseries index
        # name takes precedence
        zone_gser.index.name = "name"
        za = pls.ZonalAnalysis(self.landscape_fp, zone_gser)
        self.assertEqual(za.landscape_ser.index.name, zone_gser.index.name)
        # test that for an named geoseries with an unnamed index, the geoseries name is
        # taken
        zone_gser.index.name = None
        za = pls.ZonalAnalysis(self.landscape_fp, zone_gser)
        self.assertEqual(za.landscape_ser.index.name, zone_gser.name)
        # test overriding the zone index
        zone_index = zone_gser.index + 1
        za = pls.ZonalAnalysis(self.landscape_fp, zone_gser, zone_index=zone_index)
        self.assertTrue(np.all(za.zone_gser.index == zone_index))
        # test overriding the zone index with a named index
        zone_index = zone_index.rename("foo")
        za = pls.ZonalAnalysis(self.landscape_fp, zone_gser, zone_index=zone_index)
        self.assertEqual(za.landscape_ser.index.name, zone_index.name)
        # test that for a list-like of shapely geometries, the CRS of the landscape is
        # taken
        zones = list(zone_gser)
        za = pls.ZonalAnalysis(self.landscape_fp, zones)
        self.assertEqual(za.zone_gser.crs, self.landscape_crs)

        # now test the GeoDataFrame and geopandas file
        zone_index_col = "GMDNAME"
        for zones in self.zones_fp, self.zone_gdf:
            # test init
            za = pls.ZonalAnalysis(self.landscape_fp, zones)
            self.assertLessEqual(len(za), len(self.zone_gdf))
            self.assertTrue(
                np.all(np.isin(za.landscape_ser.index, self.zone_gdf.index))
            )
            # test that we can set a column as the zone index
            za = pls.ZonalAnalysis(self.landscape_fp, zones, zone_index=zone_index_col)
            self.assertTrue(
                np.all(
                    np.isin(
                        za.zone_gser.index,
                        self.zone_gdf[zone_index_col],
                    )
                )
            )

        # test that we can still pass a raster labelled array
        # test that the landscape index name and values are consistent with the provided
        # `label_arr`
        za = pls.ZonalAnalysis(self.landscape_fp, self.label_arr)
        self.assertEqual(za.landscape_ser.index.name, "zone")
        # the number of zones must be equal to the number of unique labels (excluding
        # the nodata value)
        zone_nodata = 0
        zones = list(set(np.unique(self.label_arr)).difference({zone_nodata}))
        self.assertEqual(len(za), len(zones))
        # test that Landscape instances are automatically built
        for landscape in za.landscape_ser:
            self.assertIsInstance(landscape, pls.Landscape)
        # test that the zone index corresponds to the zone labels
        self.assertTrue(np.all(np.isin(zones, za.zone_gser.index)))
        # test that we can override the zone index
        zone_index = pd.Series(range(1, len(zones) + 1), name="foo")
        za = pls.ZonalAnalysis(self.landscape_fp, self.label_arr, zone_index=zone_index)
        self.assertEqual(za.landscape_ser.index.name, "foo")
        self.assertTrue(np.all(za.zone_gser.index == zone_index))

        # test that we can override the nodata value
        zone_nodata = 255
        label_arr = np.where(self.label_arr != 0, self.label_arr, zone_nodata)
        # in this case, if we do not specify the nodata value, we will have another zone
        # (since by default, 0 are considered nodata)
        self.assertEqual(
            len(pls.ZonalAnalysis(self.landscape_fp, label_arr)), len(zones) + 1
        )
        self.assertEqual(
            len(
                pls.ZonalAnalysis(self.landscape_fp, label_arr, zone_nodata=zone_nodata)
            ),
            len(zones),
        )

        # test the `neighborhood_rule` argument
        neighborhood_rule = "4"
        # test that if provided, the value is passed to each landscape
        for ls in pls.ZonalAnalysis(
            self.landscape_fp,
            zone_gser,
            neighborhood_rule=neighborhood_rule,
        ).landscape_ser:
            self.assertEqual(ls.neighborhood_rule, neighborhood_rule)
        # test that if not provided, the default value is taken
        for ls in pls.ZonalAnalysis(self.landscape_fp, self.label_arr).landscape_ser:
            self.assertEqual(
                ls.neighborhood_rule, pls.settings.DEFAULT_NEIGHBORHOOD_RULE
            )

    def test_zonal_plot_metrics(self):
        za = pls.ZonalAnalysis(self.landscape_fp, self.zone_gdf)

        # test for `None` (landscape-level) and an existing class (class-level)
        for class_val in [None, za.present_classes[0]]:
            # test that the x data of the line corresponds to the attribute values,
            # i.e., the zone ids
            self.assertTrue(
                np.all(
                    za.plot_metric("patch_density", class_val=class_val)
                    .lines[0]
                    .get_xdata()
                    == za.zone_gser.index
                )
            )

    def test_compute_zonal_statistics_gdf(self):
        za = pls.ZonalAnalysis(self.landscape_fp, self.zone_gdf)

        # test that the gdf has the proper shape (number of zones, number of metrics +
        # geometry column)
        for class_val in [None, za.present_classes[0]]:
            metrics = ["patch_density"]
            zs_gdf = za.compute_zonal_statistics_gdf(
                metrics=metrics, class_val=class_val
            )
            self.assertEqual(zs_gdf.shape, (len(self.zone_gdf), len(metrics) + 1))
            # test that the crs is set correctly
            self.assertEqual(zs_gdf.crs, self.zone_gdf.crs)
            # test that the geometry column is not None
            self.assertFalse(zs_gdf.geometry.isna().any())

            # test that the zonal statistics when excluding boundaries should be less or
            # equal than including them
            metric = "total_edge"
            metric_kwargs = {"count_boundary": True}
            self.assertLessEqual(
                za.compute_zonal_statistics_gdf(metrics=[metric], class_val=class_val)[
                    metric
                ].sum(),
                za.compute_zonal_statistics_gdf(
                    metrics=[metric],
                    class_val=class_val,
                    metrics_kwargs={metric: metric_kwargs},
                )[metric].sum(),
            )

    def test_buffer_init(self):
        naive_gser = gpd.GeoSeries([self.geom])
        gser = gpd.GeoSeries([self.geom], crs=geom_crs)
        naive_gdf = gpd.GeoDataFrame(geometry=naive_gser)
        gdf = gpd.GeoDataFrame(geometry=gser)

        # test that we cannot init from a shapely geometry without providing its crs
        self.assertRaises(
            ValueError,
            pls.BufferAnalysis,
            self.landscape_fp,
            self.geom,
            self.buffer_dists,
        )

        # test that we can properly instantiate it from:
        # 1. a landscape filepath, shapely geometry, and its crs
        # 2. a landscape filepath, naive geopandas GeoSeries/GeoDataFrame (with no crs
        #    set) and its crs
        # 3. a landscape filepath, geopandas GeoSeries/GeoDataFrame with crs set
        # 4. a landscape filepath, geopandas GeoSeries/GeoDataFrame with crs set and a
        #    crs (which will override the crs of the GeoSeries)
        for ba in [
            pls.BufferAnalysis(
                self.landscape_fp,
                self.geom,
                self.buffer_dists,
                base_geom_crs=geom_crs,
            ),
            pls.BufferAnalysis(
                self.landscape_fp,
                naive_gser,
                self.buffer_dists,
                base_geom_crs=geom_crs,
            ),
            pls.BufferAnalysis(self.landscape_fp, gser, self.buffer_dists),
            pls.BufferAnalysis(
                self.landscape_fp,
                gser,
                self.buffer_dists,
                base_geom_crs=geom_crs,
            ),
            pls.BufferAnalysis(
                self.landscape_fp,
                naive_gdf,
                self.buffer_dists,
                base_geom_crs=geom_crs,
            ),
            pls.BufferAnalysis(self.landscape_fp, gdf, self.buffer_dists),
            pls.BufferAnalysis(
                self.landscape_fp,
                gdf,
                self.buffer_dists,
                base_geom_crs=geom_crs,
            ),
        ]:
            self.assertEqual(ba.landscape_ser.index.name, "buffer_dist")
            self.assertEqual(len(ba), len(ba.zone_gser))

        # test that we cannot instantiate a `BufferAnalysis` with `buffer_rings=True` if
        # `base_mask` is a polygon or a GeoSeries containing a polygon
        polygon = self.geom.buffer(1000)  # this will return a polygon instance
        self.assertRaises(
            ValueError,
            pls.BufferAnalysis,
            self.landscape_fp,
            polygon,
            self.buffer_dists,
            buffer_rings=True,
            base_geom_crs=geom_crs,
        )
        polygon_gser = gpd.GeoSeries([polygon], crs=geom_crs)
        self.assertRaises(
            ValueError,
            pls.BufferAnalysis,
            self.landscape_fp,
            polygon_gser,
            self.buffer_dists,
            buffer_rings=True,
        )

        # test that otherwise, buffer rings are properly instantiated
        ba_rings = pls.BufferAnalysis(
            self.landscape_fp, gser, self.buffer_dists, buffer_rings=True
        )
        # the `buffer_dists` attribute must be a string of the form '{r}-{R}' where r
        # and R respectively represent the smaller and larger radius that compose each
        # ring
        for buffer_ring_str in ba_rings.zone_gser.index:
            self.assertIn("-", buffer_ring_str)

        # compare it with the default instance (with the argument `buffer_rings=False`)
        # that does not consider rings but cumulatively considers the inner areas in
        # each zone. The first zone will in fact be the same in both cases (the region
        # that goes from 0 to the first item of `self.buffer_dists`), but the successive
        # zones will be always larger in the default instance (since they will have the
        # surface of the corresponding ring plus the surface of the inner region that is
        # excluded when `buffer_rings=True`)
        ba = pls.BufferAnalysis(self.landscape_fp, gser, self.buffer_dists)
        for zone_gser, ring_zone_gser in zip(ba.zone_gser, ba_rings.zone_gser):
            self.assertGreaterEqual(zone_gser.area, ring_zone_gser.area)

    def test_grid_init(self):
        # test init by number of zone rows/cols
        num_zone_rows, num_zone_cols = 10, 10
        zga = pls.ZonalGridAnalysis(
            self.landscape_fp,
            num_zone_rows=num_zone_rows,
            num_zone_cols=num_zone_cols,
        )
        # there are at most as many zones as num_zone_rows * num_zone_cols, because
        # cells with no valid data are excluded from `zone_gser`
        self.assertLessEqual(len(zga.zone_gser), num_zone_rows * num_zone_cols)

        # test init by zone width/height
        zone_width, zone_height = 1000, 1000
        zga = pls.ZonalGridAnalysis(
            self.landscape_fp,
            zone_width=zone_width,
            zone_height=zone_height,
        )

        # test that init must provide one arg for each dimension (i.e,
        # `num_zone_cols`/`zone_width` and `num_zone_rows`/`zone_height`)
        for kws in [
            {},
            {
                "zone_height": zone_height,
                "num_zone_rows": num_zone_rows,
            },
            {
                "zone_width": zone_width,
                "num_zone_cols": num_zone_cols,
            },
        ]:
            with self.assertRaises(ValueError) as cm:
                zga = pls.ZonalGridAnalysis(self.landscape_fp, **kws)
            self.assertIn("must be provided", str(cm.exception))

        # test the `neighborhood_rule` argument
        neighborhood_rule = "4"
        # test that the value is passed to each landscape
        for ls in pls.ZonalGridAnalysis(
            self.landscape_fp,
            zone_width=zone_width,
            zone_height=zone_height,
            neighborhood_rule=neighborhood_rule,
        ).landscape_ser:
            self.assertEqual(ls.neighborhood_rule, neighborhood_rule)
        # test that if not provided, the default value is taken
        for ls in pls.ZonalGridAnalysis(
            self.landscape_fp,
            zone_width=zone_width,
            zone_height=zone_height,
        ).landscape_ser:
            self.assertEqual(
                ls.neighborhood_rule, pls.settings.DEFAULT_NEIGHBORHOOD_RULE
            )


class TestSpatioTemporalZonalAnalysis(unittest.TestCase):
    def setUp(self):
        self.landscape_fps = [
            path.join(tests_data_dir, "ls250_06.tif"),
            path.join(tests_data_dir, "ls250_12.tif"),
        ]
        self.dates = [2006, 2012]
        # zonal
        self.zone_gser = pls.ZonalAnalysis(
            self.landscape_fps[0],
            np.load(path.join(tests_data_dir, "label_arr.npy"), allow_pickle=True),
        ).zone_gser
        # buffer
        self.base_geom = gpd.GeoSeries(
            [geometry.Point(6.6327025, 46.5218269)], crs=geom_crs
        )
        self.buffer_dists = [10000, 15000, 20000]
        # grid
        self.num_zone_rows, self.num_zone_cols = 10, 10
        # self.zone_width, self.zone_height = 1000, 1000
        self.init_combinations = zip(
            [
                pls.SpatioTemporalZonalAnalysis,
                pls.SpatioTemporalBufferAnalysis,
                pls.SpatioTemporalZonalGridAnalysis,
            ],
            [[self.zone_gser], [self.base_geom, self.buffer_dists], []],
            [
                {},
                {},
                {
                    "num_zone_cols": self.num_zone_cols,
                    "num_zone_rows": self.num_zone_rows,
                },
            ],
            ["zone", "buffer_dist", "grid_cell"],
        )

    def test_init(self):
        # we will just test the base init, the rest of functionalities have already been
        # tested above (in `TestSpatioTemporalAnalysis` and `TestZonalAnalysis`)
        for _class, init_args, init_kwargs, attr_name in self.init_combinations:
            # test zones and dates
            stza = _class(
                self.landscape_fps, *init_args, dates=self.dates, **init_kwargs
            )
            # test that we have the same zone labels in `zone_gser` and `landscape_ser`
            self.assertTrue(
                (
                    stza.zone_gser.index
                    == stza.landscape_ser.index.get_level_values(
                        stza.zone_gser.index.name
                    ).unique()
                ).all()
            )
            # test the `zone_gser` index name
            self.assertEqual(stza.zone_gser.index.name, attr_name)
            # test that we have the same date labels in `dates` and `landscape_ser`
            self.assertTrue(
                (
                    self.dates
                    == stza.landscape_ser.index.get_level_values("date").unique()
                ).all()
            )

            # test the `neighborhood_rule` argument
            # test that if not provided, the default value is taken
            for ls in stza.landscape_ser:
                self.assertEqual(
                    ls.neighborhood_rule,
                    pls.settings.DEFAULT_NEIGHBORHOOD_RULE,
                )
            neighborhood_rule = "4"
            # test that if provided, the value is passed to each landscape
            stza = _class(
                self.landscape_fps,
                *init_args,
                neighborhood_rule=neighborhood_rule,
                **init_kwargs,
            )
            for ls in stza.landscape_ser:
                self.assertEqual(ls.neighborhood_rule, neighborhood_rule)

    def test_dataframes(self):
        for _class, init_args, init_kwargs, _ in self.init_combinations:
            stza = _class(
                self.landscape_fps, *init_args, dates=self.dates, **init_kwargs
            )
            zone_index = stza.zone_gser.index
            dates = stza.landscape_ser.index.get_level_values("date").unique()

            # test that the data frames that result from `compute_class_metrics_df` and
            # `compute_landscape_metrics_df` are well constructed, i.e., they are a
            # subset of all class/zone/date combinations
            class_metrics_df = stza.compute_class_metrics_df()
            self.assertTrue(
                class_metrics_df.index.isin(
                    pd.MultiIndex.from_product(
                        [stza.present_classes, zone_index, dates]
                    )
                ).all()
            )
            landscape_metrics_df = stza.compute_landscape_metrics_df()
            self.assertTrue(
                landscape_metrics_df.index.isin(
                    pd.MultiIndex.from_product([zone_index, dates])
                ).all()
            )

            # now test the same but with an analysis that only considers a subset of
            # metrics and a subset of classes
            metrics = ["total_area", "edge_density", "proportion_of_landscape"]
            classes = stza.present_classes[:2]

            class_metrics_df = stza.compute_class_metrics_df(
                metrics=metrics, classes=classes
            )
            self.assertTrue(
                class_metrics_df.index.isin(
                    pd.MultiIndex.from_product([classes, zone_index, dates])
                ).all()
            )
            # 'proportion_of_landscape' cannot be computed at the landscape level (TODO:
            # test for that elsewhere)
            landscape_metrics = metrics[:2]
            landscape_metrics_df = stza.compute_landscape_metrics_df(
                metrics=landscape_metrics
            )
            self.assertTrue(
                landscape_metrics_df.index.isin(
                    pd.MultiIndex.from_product([zone_index, dates])
                ).all()
            )

    def test_compute_zonal_statistics_gdf(self):
        for _class, init_args, init_kwargs, _ in self.init_combinations:
            stza = _class(
                self.landscape_fps, *init_args, dates=self.dates, **init_kwargs
            )
            # test that the gdf has the proper shape (number of zones, number of metrics
            # + geometry column)
            for class_val in [None, stza.present_classes[0]]:
                metrics = ["patch_density"]
                zs_gdf = stza.compute_zonal_statistics_gdf(
                    metrics=metrics, class_val=class_val
                )
                self.assertLessEqual(
                    zs_gdf.shape[0],
                    len(stza.zone_gser),
                )
                self.assertEqual(zs_gdf.shape[1], len(metrics) * len(self.dates) + 1)
                # test that the crs is set correctly
                self.assertEqual(zs_gdf.crs, self.zone_gser.crs)
                # test that the geometry column is not None
                self.assertFalse(zs_gdf.geometry.isna().any())

                # test that the zonal statistics when excluding boundaries should be
                # less or equal than including them
                metric = "total_edge"
                metric_kwargs = {"count_boundary": True}
                self.assertLessEqual(
                    stza.compute_zonal_statistics_gdf(
                        metrics=[metric], class_val=class_val
                    )[metric]
                    .sum()
                    .sum(),
                    stza.compute_zonal_statistics_gdf(
                        metrics=[metric],
                        class_val=class_val,
                        metrics_kwargs={metric: metric_kwargs},
                    )[metric]
                    .sum()
                    .sum(),
                )

    def test_plot_metric(self):
        for _class, init_args, init_kwargs, _ in self.init_combinations:
            stza = _class(
                self.landscape_fps, *init_args, dates=self.dates, **init_kwargs
            )
            # test for `None` (landscape-level) and an existing class (class-level)
            metric = "patch_density"
            for class_val in [None, stza.present_classes[0]]:
                ax = stza.plot_metric(metric, class_val=class_val)
                # test that there is a line for each zone with that metric
                if class_val is None:
                    metric_ser = stza.compute_landscape_metrics_df(metrics=[metric])[
                        metric
                    ]
                else:
                    metric_ser = stza.compute_class_metrics_df(
                        metrics=[metric], classes=[class_val]
                    )[metric]
                num_lines = len(
                    metric_ser.index.get_level_values(
                        stza.zone_gser.index.name
                    ).unique()
                )
                self.assertEqual(
                    len(ax.lines),
                    num_lines,
                )
                # test that there is a legend label for each zone
                _, labels = ax.get_legend_handles_labels()
                self.assertEqual(len(labels), num_lines)

    def test_plot_landscapes(self):
        for _class, init_args, init_kwargs, _ in self.init_combinations:
            stza = _class(
                self.landscape_fps, *init_args, dates=self.dates, **init_kwargs
            )
            dates = stza.landscape_ser.index.get_level_values("date").unique()

            fig = stza.plot_landscapes()

            # there must be one column for each buffer distance and one row for each
            # date
            self.assertEqual(len(fig.axes), len(stza.zone_gser) * len(dates))

            # returned axes must be instances of matplotlib axes
            for ax in fig.axes:
                self.assertIsInstance(ax, plt.Axes)

            # test that by default, the dimensions of the resulting will come from
            # matplotlib's settings
            rc_figwidth, rc_figheight = plt.rcParams["figure.figsize"]
            figwidth, figheight = fig.get_size_inches()
            # the actual `figwidth` must be `len(stba.buffer_dists) * rc_figwidth` and
            # `figheight` must be `len(stba.dates) * rc_figheight`
            self.assertAlmostEqual(figwidth, len(stza.zone_gser) * rc_figwidth)
            self.assertAlmostEqual(figheight, len(dates) * rc_figheight)
            # if instead, we customize the figure size, the dimensions of the resulting
            # figure must be the customized ones
            custom_figsize = (10, 10)
            fig = stza.plot_landscapes(subplots_kwargs={"figsize": custom_figsize})
            figwidth, figheight = fig.get_size_inches()
            self.assertAlmostEqual(custom_figsize[0], figwidth)
            self.assertAlmostEqual(custom_figsize[1], figheight)

            # first row has the date as title
            for date, ax in zip(dates, fig.axes):
                self.assertEqual(str(date), ax.get_title())
            # first column has the buffer distance as `ylabel`
            for zone, i in zip(
                stza.zone_gser.index, range(0, len(fig.axes), len(dates))
            ):
                self.assertEqual(str(zone), fig.axes[i].get_ylabel())


class TestSpatialSignatureAnalysis(unittest.TestCase):
    def setUp(self):
        self.landscape_fps = [
            path.join(tests_data_dir, "ls250_06.tif"),
            path.join(tests_data_dir, "ls250_12.tif"),
        ]
        dates = [2006, 2012]
        self.sta = pls.SpatioTemporalAnalysis(self.landscape_fps, dates=dates)
        # zones_fp = path.join(tests_data_dir, "gmb-lausanne.gpkg")
        # use label_arr instead of zones_fp to have at least 2 zones
        label_arr = np.load(
            path.join(tests_data_dir, "label_arr.npy"), allow_pickle=True
        )
        self.za = pls.ZonalAnalysis(self.landscape_fps[0], label_arr)
        self.stza = pls.SpatioTemporalZonalAnalysis(self.landscape_fps, label_arr)

        self.class_metrics = ["proportion_of_landscape"]
        self.landscape_metrics = ["edge_density"]

        # use to test all the methods except `test_metrics_df` (which tests init
        # combinations). Note that we only use the first class to avoid nan values so
        # that we can instantiate without an imputer
        self.sta_ssa = pls.SpatialSignatureAnalysis(
            self.sta,
            class_metrics=self.class_metrics,
            classes=self.sta.present_classes[:1],
            landscape_metrics=self.landscape_metrics,
        )

        # use this to test with nans
        self.zonal_nan_ssa = pls.SpatialSignatureAnalysis(
            self.za,
            class_metrics=self.class_metrics,
            classes=self.za.present_classes[:1],
            landscape_metrics=self.landscape_metrics,
        )

    #     self.tmp_dir = path.join(tests_dir, "tmp")
    #     os.mkdir(self.tmp_dir)

    # def tearDown(self):
    #     shutil.rmtree(self.tmp_dir)

    def test_metrics_df(self):
        for landscapes in [self.sta, self.za, self.stza]:
            classes = landscapes.present_classes
            classes_subset = classes[1:]
            # test that we can instantiate the class with any of the multilandscape
            # classes
            for kwargs, n_cols in [
                (
                    {"class_metrics": self.class_metrics},
                    len(classes) * len(self.class_metrics),
                ),
                (
                    {"class_metrics": self.class_metrics, "classes": classes_subset},
                    len(classes_subset) * len(self.class_metrics),
                ),
                (
                    {"class_metrics": self.class_metrics, "class_metrics_fillna": 0},
                    len(classes) * len(self.class_metrics),
                ),
                (
                    {
                        "class_metrics": self.class_metrics,
                        "class_metrics_kwargs": {
                            "proportion_of_landscape": {"percent": True}
                        },
                    },
                    len(classes) * len(self.class_metrics),
                ),
                (
                    {"landscape_metrics": self.landscape_metrics},
                    len(self.landscape_metrics),
                ),
                (
                    {
                        "landscape_metrics": self.landscape_metrics,
                        "landscape_metrics_kwargs": {
                            "edge_density": {"count_boundary": True}
                        },
                    },
                    len(self.landscape_metrics),
                ),
                (
                    {
                        "class_metrics": self.class_metrics,
                        "landscape_metrics": self.landscape_metrics,
                    },
                    len(classes) * len(self.class_metrics)
                    + len(self.landscape_metrics),
                ),
            ]:
                ssa = pls.SpatialSignatureAnalysis(landscapes, **kwargs)

                # test that the metrics_df index is the same as that from
                # landscapes.landscape_ser
                assert_index_equal(ssa.metrics_df.index, landscapes.landscape_ser.index)
                # test that the shape of the metrics data frame is correct
                self.assertEqual(ssa.metrics_df.shape[1], n_cols)

    def test_decompose(self):
        for kwargs in [
            {},
            {"preprocessor": preprocessing.StandardScaler},
            {
                "preprocessor": preprocessing.StandardScaler,
                "preprocessor_kwargs": {"with_mean": False},
            },
            {"imputer": impute.SimpleImputer},
            {"imputer": impute.SimpleImputer, "imputer_kwargs": {"strategy": "mean"}},
            {"decomposer": decomposition.PCA},
            # note that the decomposer kwargs are passed directly to the `decompose`
            # method
            {"decomposer": decomposition.PCA, "n_components": 2},
        ]:
            components_df, decompose_model = self.sta_ssa.decompose(**kwargs)

            # test that decompose_model has the fit, transform and fit_transform methods
            for method_name in ["fit", "transform", "fit_transform"]:
                method = getattr(decompose_model, method_name)
                self.assertTrue(callable(method))

            # test that decompose_model has the `components_` and `n_components_`
            # attributes
            for attr_name in ["components_", "n_components_"]:
                self.assertTrue(hasattr(decompose_model, attr_name))

            # test that the shape of the components_df is correct
            self.assertLessEqual(
                components_df.shape[0], self.sta_ssa.metrics_df.shape[0]
            )
            self.assertEqual(components_df.shape[1], decompose_model.n_components_)

            # test loading_df
            loading_df = self.sta_ssa.get_loading_df(decompose_model)
            self.assertEqual(loading_df.shape[0], self.sta_ssa.metrics_df.shape[1])
            self.assertEqual(loading_df.shape[1], decompose_model.n_components_)

        # test warning when `metrics_df` has nan values
        with warnings.catch_warnings(record=True) as w:
            components_df, decompose_model = self.zonal_nan_ssa.decompose(**kwargs)
            self.assertGreater(len(w), 0)
        # if no imputer is passed, rows are dropped
        self.assertLess(components_df.shape[0], self.zonal_nan_ssa.metrics_df.shape[0])

        # if an imputer is passed, the returned number of landscapes (samples) should be
        # the same
        components_df, decompose_model = self.zonal_nan_ssa.decompose(
            imputer=impute.SimpleImputer, **kwargs
        )
        self.assertEqual(components_df.shape[0], self.zonal_nan_ssa.metrics_df.shape[0])

    def test_cgram(self):
        # test that instantiating with no args will raise a ValueError because it will
        # try to fit n_clusters > n_samples (2)
        with self.assertRaises(ValueError) as cm:
            self.sta_ssa.get_cgram()
        self.assertIn("should be >=", str(cm.exception))

        # test different init kwargs but always with custom k_range
        base_cgram_kwargs = {"k_range": [2]}

        for kwargs in [
            {},
            {"decomposer": decomposition.PCA},
            {"decomposer": decomposition.PCA, "decomposer_kwargs": {"n_components": 2}},
            {"preprocessor": preprocessing.StandardScaler},
            {
                "preprocessor": preprocessing.StandardScaler,
                "preprocessor_kwargs": {"with_mean": False},
            },
            {"imputer": impute.SimpleImputer},
            {"imputer": impute.SimpleImputer, "imputer_kwargs": {"strategy": "mean"}},
            # custom Clustergram kwargs (besides k_range)
            {"method": "gmm"},
        ]:
            _kwargs = dict(kwargs, **base_cgram_kwargs)
            cgram = self.sta_ssa.get_cgram(**_kwargs)

        # test warning when `metrics_df` has nan values
        with warnings.catch_warnings(record=True) as w:
            cgram = self.zonal_nan_ssa.get_cgram(**base_cgram_kwargs)
            self.assertGreater(len(w), 0)
        # if no imputer is passed, rows are dropped
        self.assertLess(cgram.data.shape[0], self.zonal_nan_ssa.metrics_df.shape[0])

        # if an imputer is passed, the returned number of landscapes (samples) should be
        # the same
        cgram = self.zonal_nan_ssa.get_cgram(
            imputer=impute.SimpleImputer, **base_cgram_kwargs
        )
        self.assertEqual(cgram.data.shape[0], self.zonal_nan_ssa.metrics_df.shape[0])

        # use the latest cgram (with `self.zonal_nan_ssa`) to test the plotting methods
        n_clusters = 2
        # test scatterplot_cluster_metrics
        scatterplot_args = [
            cgram,
            n_clusters,
            self.zonal_nan_ssa.metrics_df.columns[0],
            self.zonal_nan_ssa.metrics_df.columns[1],
        ]
        for kwargs in [
            {"palette_name": "Set1"},
            {"center_plot_kwargs": {"marker": "v"}},
        ]:
            ax = self.zonal_nan_ssa.scatterplot_cluster_metrics(
                *scatterplot_args, **kwargs
            )
            # test that there are as many legend entries as clusters
            self.assertEqual(len(ax.get_legend().get_texts()), n_clusters)
        # test plot in given axis
        fig, ax = plt.subplots()
        self.zonal_nan_ssa.scatterplot_cluster_metrics(*scatterplot_args, ax=ax)
        self.assertEqual(len(ax.get_legend().get_texts()), n_clusters)
        # test no legend
        ax = self.zonal_nan_ssa.scatterplot_cluster_metrics(
            *scatterplot_args, legend=False
        )
        self.assertIsNone(ax.get_legend())

        # test plot_cluster_landscapes
        plot_landscapes_kwargs = {"n_cluster_landscapes": 1, "n_cols": 1}
        for kwargs in [
            {},
            {"sample_kwargs": {"random_state": 1}},
            {"figure_kwargs": {"figsize": (1, 1)}},
            {"subfigures_kwargs": {"wspace": 0.1}},
            {"subplots_kwargs": {"sharex": True, "sharey": True}},
            {"supylabel_kwargs": {"fontweight": "bold"}},
            {"cmap": "viridis"},
        ]:
            _kwargs = dict(kwargs, **plot_landscapes_kwargs)
            self.zonal_nan_ssa.plot_cluster_landscapes(cgram, n_clusters, **_kwargs)

        # test plot_cluster_zones
        for kwargs in [{}, {"categorical": False}, {"cmap": "viridis"}]:
            ax = self.zonal_nan_ssa.plot_cluster_zones(cgram, n_clusters, **kwargs)
            # self.assertEqual(len(ax.get_legend().get_texts()), n_clusters)
        # test plot in given axis
        fig, ax = plt.subplots()
        self.zonal_nan_ssa.plot_cluster_zones(cgram, n_clusters, ax=ax)
        self.assertEqual(len(ax.get_legend().get_texts()), n_clusters)
        # test no legend
        ax = self.zonal_nan_ssa.plot_cluster_zones(cgram, n_clusters, legend=False)
        self.assertIsNone(ax.get_legend())

        # test that a warning is issued when calling `plot_cluster_zones` if the
        # SpatialSignatureAnalysis is not initialized with a zonal analysis, and None is
        # returned
        ssa = pls.SpatialSignatureAnalysis(
            self.sta,
            class_metrics=self.class_metrics,
            classes=self.sta.present_classes[:1],
            landscape_metrics=self.landscape_metrics,
        )
        cgram = ssa.get_cgram(**base_cgram_kwargs)
        with warnings.catch_warnings(record=True) as w:
            ax = ssa.plot_cluster_zones(cgram, n_clusters)
            self.assertGreater(len(w), 0)

        self.assertIsNone(ax)

        # TODO: test plotting zones in a spatiotemporalzonalanalysis
        # test that a figure is returned, with as many axes as dates
        ssa = pls.SpatialSignatureAnalysis(
            self.stza,
            class_metrics=self.class_metrics,
            classes=self.sta.present_classes[:1],
            landscape_metrics=self.landscape_metrics,
        )
        cgram = ssa.get_cgram(**base_cgram_kwargs)
        fig = ssa.plot_cluster_zones(cgram, n_clusters)
        self.assertIsInstance(fig, plt.Figure)
        self.assertEqual(len(fig.axes), len(self.stza.dates))



import numpy as np
cimport cython
cimport numpy as np

# define a C-level array type
ctypedef np.uint32_t DTYPE_t  # data type for elements


@cython.boundscheck(False)  # deactivate bounds checking
@cython.wraparound(False)  # deactivate negative indexing.
def compute_adjacency_arr(np.ndarray[DTYPE_t, ndim=2] padded_arr, int num_classes):
    """Compute adjacency array using memoryviews."""

    # prepare adjacency array with memoryviews
    cdef int num_cols_adjacency = num_classes + 1
    cdef int num_cols_pixel = padded_arr.shape[1]
    cdef np.ndarray[DTYPE_t, ndim=1] horizontal_adjacency_arr = np.zeros(
        num_cols_adjacency * num_cols_adjacency, dtype=np.uint32)
    cdef np.ndarray[DTYPE_t, ndim=1] vertical_adjacency_arr = np.zeros(
        num_cols_adjacency * num_cols_adjacency, dtype=np.uint32)

    # memoryviews for efficient access
    cdef DTYPE_t[:] flat_arr = padded_arr.ravel()
    cdef DTYPE_t[:] horizontal_adj_mem = horizontal_adjacency_arr
    cdef DTYPE_t[:] vertical_adj_mem = vertical_adjacency_arr

    # define neighbor offsets
    cdef int horizontal_neighbors[2]
    horizontal_neighbors[0] = 1
    horizontal_neighbors[1] = -1
    cdef int vertical_neighbors[2]
    vertical_neighbors[0] = num_cols_pixel
    vertical_neighbors[1] = -num_cols_pixel

    cdef int start = num_cols_pixel + 1
    cdef int end = flat_arr.shape[0] - start

    cdef int i, class_i, neighbor, neighbor_idx, target_idx

    for i in range(start, end):
        class_i = flat_arr[i]

        # update horizontal adjacency counts
        for neighbor in horizontal_neighbors:
            neighbor_idx = i + neighbor
            target_idx = class_i + num_cols_adjacency * flat_arr[neighbor_idx]
            horizontal_adj_mem[target_idx] += 1

        # update vertical adjacency counts
        for neighbor in vertical_neighbors:
            neighbor_idx = i + neighbor
            target_idx = class_i + num_cols_adjacency * flat_arr[neighbor_idx]
            vertical_adj_mem[target_idx] += 1

    # return the result as a reshaped numpy array
    return np.stack((horizontal_adjacency_arr, vertical_adjacency_arr)).reshape(
        (2, num_cols_adjacency, num_cols_adjacency)
    )
